"""
Created on Fri Dec 20 12:33:30 2019
Modified on Tue Jun 21 12:32:00 2022

@author: VICTOR
"""
import numpy as np


def signed_r2(class1, class2, signed=True, axis=0):
    """ This function computes the basic form of the squared point biserial
    correlation coefficient (r2-value).

    Parameters
    ---------
    class1: list or numpy.ndarray
        Data that belongs to the first class
    class1: list or numpy.ndarray
        Data that belongs to the second class
    signed: bool (Optional, default=True)
        Controls if the sign should be mantained.
    axis: int (Optional, default=0)
        Dimension along which the r2-value is computed. Therefore,
        if class1 and class2 has dimensions of [observations x samples]
        and dim=0, the r2-value will have dimensions [1 x samples].

    Returns
    -------
    r2: numpy.ndarray
        (Signed) r2-value.
    """
    # Length of each class
    N1 = class1.shape[axis]
    N2 = class2.shape[axis]
    
    # Pre-computation
    all_data = np.concatenate((class1,class2), axis=axis)
    v = np.var(all_data, axis=axis)
    m_diff = np.mean(class1, axis=axis) - np.mean(class2, axis=axis)
    
    # Compute the sign if required
    sign = 1
    if signed:
        sign = np.sign(m_diff)
        sign[sign == 0] = 1
    
    # Final r2 value
    r2 = sign*N1*N2*np.power(m_diff, 2)/(v*np.power(N1+N2, 2))
    return r2



import csv, math, os

# Synonyms
eeg_channel_synonyms = {
    'I1': 'O9',
    'I2': 'O10',
    'O9': 'I1',
    'O10': 'I2',
    'T3': 'T7',
    'T4': 'T8',
    'T7': 'T3',
    'T8': 'T4',
    'T5': 'P7',
    'T6': 'P8'
}


def get_standard_montage(standard, dim, coord_system):
    """Loads a standard from the ones included in MEDUSA
    """
    # Check errors
    assert standard == '10-20' or standard == '10-10' or standard == '10-05', \
        'Incorrect input on standard parameter. Possible values {"10-20", ' \
        '"10-10", "10-05"}'
    assert dim == '2D' or dim == '3D', \
        'Incorrect input on dim parameter. Possible values {"2D", "3D"}'
    assert coord_system == 'cartesian' or coord_system == 'spherical', \
        'Incorrect input on coord_system parameter. Possible values ' \
        '{"cartesian", "spherical"}'

    folder = os.path.dirname(__file__)
    montage = read_montage_file(
        path='%s/eeg_standard_%s_%s.tsv' % (folder, standard, dim),
        file_format='tsv', dim=dim, coord_system='cartesian')

    if coord_system == 'spherical':
        montage = switch_coordinates_system(montage, dim, 'cartesian')

    return montage


def switch_coordinates_system(montage, dim, coord_system):
    """Switch the coordinates system of the montage. If the the coordinates are
     cartesian, it switches to spherical and viceversa

    Parameters
    ----------
    montage: dict
        Montage data
    dim: str {'2D', '3D'}
        Current dimensions of the montage
    coord_system: str {'cartesian', 'spherical'}
        Current coordinates system
    """
    # Make copy to not modify the original dict
    new_montage = dict()
    # Make conversion
    for label, coordinates in montage.items():
        if coord_system == 'cartesian':
            cha_data = dict()
            if dim == '2D':
                cha_data['r'] = math.sqrt(
                    math.pow(montage[label]['x'], 2) +
                    math.pow(montage[label]['y'], 2))
                cha_data['theta'] = math.atan2(montage[label]['y'],
                                               montage[label]['x'])
            elif dim == '3D':
                cha_data['r'] = math.sqrt(
                    math.pow(montage[label]['x'], 2) +
                    math.pow(montage[label]['y'], 2) +
                    math.pow(montage[label]['z'], 2))
                cha_data['theta'] = math.atan2(
                    math.sqrt(math.pow(montage[label]['x'], 2) +
                              math.pow(montage[label]['y'], 2)),
                    montage[label]['z']
                )
                cha_data['phi'] = math.atan2(montage[label]['y'],
                                             montage[label]['x'])
            else:
                raise ValueError('Unknown number of dimensions')
        else:
            raise NotImplemented
        new_montage[label] = cha_data

    return new_montage


def read_montage_file(path, file_format, dim, coord_system):
    """This function reads a montage file.

    Parameters
    ----------
    path: str
        Path of the file
    file_format: str
        File format. Current accepted file formats:
            - csv (.csv): Comma separated values. Example:
                Format csv (cartesian, 2D):
                    label, x, y
                    C3, -0.3249, 0.0000
                    C4, 0.3249, 0.0000
                    Cz, 0.0000, 0.0000
                    ..., ..., ...
            - tsv (.tsv):  Tab separated values. Example:
                Format tsv (cartesian, 2D):
                    label	    x	        y
                    C3	        -0.3249	    0.0000
                    C4	        0.3249	    0.0000
                    Cz	        0.0000	    0.0000
                    ...         ...         ...
            - eeglab (.xyz): EEGLAB toolbox format. Example:
                Format eeglab (cartesian, 3D):
                    1	-3.277285	14.159082	-0.441064	AG001
                    2	-5.717206	13.304894	0.167594	AG002
                    3	-7.723112	12.035844	0.923781	AG003
                    ...         ...         ...
            - brainstorm (.txt): Brainstorm toolbox format. Example:
                Format brainstorm (cartesian, 3D):
                    C3	-4.025389	68.267399	112.962637
                    C4	-3.409963	-70.131718	111.811990
                    O1	-86.474171	34.971167	48.738844
                    ...         ...         ...
    dim: str
        Dimensions. Must be '2D' or '3D'
    coord_system: str
        For 2D, "cartesian" or "polar". For 3D "cartesian" or "spherical"
    """
    # ACCEPTED PARAMSN
    file_formats = ['csv', 'tsv', 'eeglab', 'brainstorm']
    dims = ['2D', '3D']
    coord_systems = ['cartesian', 'spherical']

    # CHECK ERRORS
    if file_format not in file_formats:
        raise ValueError("Parameter file_format must be one of %s"
                         % file_formats)
    if dim not in dims:
        raise ValueError("Parameter dim must be one of %s" % dims)
    if coord_system not in coord_systems:
        raise ValueError("Parameter coord_system must be one of %s"
                         % coord_systems)

    # LOAD COORDINATES FILES
    if file_format == 'csv':
        montage = __read_txt_file(path, coord_system, dim, delimiter=',')
    elif file_format == 'tsv':
        montage = __read_txt_file(path, coord_system, dim, delimiter='\t')
    elif file_format == 'eeglab':
        montage = __read_eeglab_file(path, dim)
    elif file_format == 'brainstorm':
        montage = __read_brainstorm_file(path, dim)
    else:
        raise ValueError("Parameter file_format must be one of %s"
                         % file_formats)
    return montage


def __read_txt_file(path, coord_system, dim, delimiter='\t'):
    data = dict()
    # 2D coordinates
    with open(path) as f:
        tsvreader = csv.reader(f, delimiter=delimiter)
        for l, line in enumerate(tsvreader):
            if dim == '2D':
                label, coord1, coord2 = line
                if l == 0:
                    # Check format errors
                    if coord_system == 'cartesian':
                        if coord1 != 'x' or coord2 != 'y':
                            raise ValueError(
                                'Invalid header. '
                                'For dim=%s, '
                                'coord_system=%s, '
                                'the header must be [label, x, y]' %
                                (dim, coord_system)
                            )
                    elif coord_system == 'spherical':
                        if coord1 != 'r' or coord2 != 'theta':
                            raise ValueError(
                                'Invalid header. '
                                'For dim=%s, '
                                'coord_system=%s, '
                                'the header must be [label, r, theta]' %
                                (dim, coord_system)
                            )
                    else:
                        raise ValueError('Invalid coordinate system')
                    continue
                else:
                    lcha = line[0].upper()
                    data[lcha] = dict()
                    if coord_system == 'cartesian':
                        data[lcha]['x'] = float(coord1)
                        data[lcha]['y'] = float(coord2)
                    elif coord_system == 'spherical':
                        data[lcha]['r'] = float(coord1)
                        data[lcha]['theta'] = float(coord2)
                    else:
                        raise ValueError('Invalid coordinate system')
            elif dim == '3D':
                label, coord1, coord2, coord3 = line
                if l == 0:
                    # Check format errors
                    if coord_system == 'cartesian':
                        if coord1 != 'x' or coord2 != 'y' or coord3 != 'z':
                            raise ValueError(
                                'Invalid header. '
                                'For dim=%s, '
                                'coord_system=%s, '
                                'the header must be [label, x, y, z]' %
                                (dim, coord_system)
                            )
                    elif coord_system == 'spherical':
                        if coord1 != 'r' or coord2 != 'theta' \
                                or coord3 != 'phi':
                            raise ValueError(
                                'Invalid header. '
                                'For dim=%s, '
                                'coord_system=%s, '
                                'the header must be [label, r, theta, phi]' %
                                (dim, coord_system)
                            )
                    else:
                        raise ValueError('Invalid coordinate system')
                    continue
                else:
                    lcha = label.upper()
                    data[lcha] = dict()
                    if coord_system == 'cartesian':
                        data[lcha]['x'] = float(coord1)
                        data[lcha]['y'] = float(coord2)
                        data[lcha]['z'] = float(coord3)
                    elif coord_system == 'spherical':
                        data[lcha]['r'] = float(coord1)
                        data[lcha]['theta'] = float(coord2)
                        data[lcha]['phi'] = float(coord3)
                    else:
                        raise ValueError('Invalid coordinate system')
    return data


def __read_eeglab_file(path, dim):
    data = dict()
    # 2D coordinates
    with open(path) as f:
        tsvreader = csv.reader(f, delimiter='\t')
        for l, line in enumerate(tsvreader):
            if dim == '2D':
                index, coord1, coord2, label = line
                # Save data
                lcha = label.upper()
                data[lcha] = dict()
                data[lcha]['x'] = float(coord1)
                data[lcha]['y'] = float(coord2)
            elif dim == '3D':
                index, coord1, coord2, coord3, label = line
                lcha = line[0].upper()
                data[lcha] = dict()
                data[lcha]['x'] = float(coord1)
                data[lcha]['y'] = float(coord2)
                data[lcha]['z'] = float(coord3)
    return data


def __read_brainstorm_file(path, dim):
    data = dict()
    # 2D coordinates
    with open(path) as f:
        tsvreader = csv.reader(f, delimiter='\t')
        for l, line in enumerate(tsvreader):
            if dim == '2D':
                label, coord1, coord2 = line
                # Save data
                lcha = label.upper()
                data[lcha] = dict()
                data[lcha]['x'] = float(coord1)
                data[lcha]['y'] = float(coord2)
            elif dim == '3D':
                label, coord1, coord2, coord3 = line
                lcha = line[0].upper()
                data[lcha] = dict()
                data[lcha]['x'] = float(coord1)
                data[lcha]['y'] = float(coord2)
                data[lcha]['z'] = float(coord3)
    return data


class ChannelNotFound(Exception):

    def __init__(self, l_cha):
        super().__init__(
            'Channel %s is not defined in the current montage' % l_cha)


class UnlocatedChannel(Exception):

    def __init__(self, l_cha):
        super().__init__(
            'Channel %s does not contain location coordinates' % l_cha)


if __name__ == '__main__':
    montage = get_standard_montage('10-20', '2D', 'spherical')
    print(montage)



"""Created on Monday February 14 09:27:14 2022

In this module you will find useful functions and classes to operate with data
recorded using motor imagery paradigms, which are widely used by the BCI
community. Enjoy!

@author: Sergio Pérez-Velasco
"""
# Built-in imports
import copy, warnings
from abc import abstractmethod

# External imports
import numpy as np
from tqdm import tqdm
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.metrics import classification_report

# Medusa imports
from medusa import IIRFilter, car
from medusa.epoching import normalize_epochs
from medusa import get_epochs_of_events, resample_epochs
from medusa import components
from medusa import meeg
from medusa.spatial_filtering import CSP, LaplacianFilter


class MIData(components.ExperimentData):
    # TODO: Check everything

    """Class with the necessary attributes to define motor imagery (MI)
    experiments. It provides compatibility with high-level functions for this
    MI paradigms in BCI module.
    """

    def __init__(self, mode, onsets, w_trial_t,
                 calibration_onset_w=None, w_preparation_t=None, w_rest_t=None,
                 mi_labels=None, mi_labels_info=None, mi_result=None,
                 paradigm_info=None, **kwargs):
        """MIData constructor

        Parameters
        ----------
        mode : str {"train"|"test"|"guided test"}
            Mode of this run.
        onsets : list or numpy.ndarray [n_stim x 1]
            Timestamp of the cue with respect to the EEG signal (just when the
            motor imagery starts).
        w_trial_t: list [start, end]
            Temporal window of the motor imagery with respect to each onset in
            ms. For example, if  w_trial_t = [500, 4000] the subject was
            performing the motor imagery task from 500ms to 4000ms after
            the onset.
        calibration_onset_w: list [start, end]
            Timestamps of the onsets regarding the calibration window,
            if exists, respect to the EEG signal.
        w_preparation_t: list [start, end]
            Temporal window of the preparation time (no motor imagery),
            if exists, with respect to each onset in ms.
        w_rest_t: list [start, end]
            Temporal window of the rest time (no motor imagery),
            if exists, with respect to each onset in ms.
        mi_labels : list or numpy.ndarray [n_mi_labels x 1]
            Only in train mode. Contains the mi labels of each stimulation,
            as many as classes in the experiment.
        mi_labels_info : dict
            Contains the description of the mi labels.
            Example:
                mi_labels_info =
                    {0: "Rest",
                    1: "Left_hand",
                    2: "Right_hand"}
        mi_result : list or numpy.ndarray [n_mi_labels x 1]
            Result of this run. Each position contains the data of the
            selected target at each trial.
        paradigm_info: dict()
            Recommended but not mandatory. Use this variable to keep the
            information regarding the different timings of the paradigm.
        kwargs : kwargs
            Custom arguments that will also be saved in the class
            (e.g., timings, calibration gaps, etc.)
        """

        # Check errors
        mode = mode.lower()
        if mode == 'train':
            if mi_labels is None:
                raise ValueError('Attribute "mi_labels" must be provided in '
                                 'train mode')

        # Standard attributes
        self.mode = mode
        self.onsets = np.array(onsets)
        self.w_trial_t = np.array(w_trial_t)
        self.calibration_onset_w = np.array(calibration_onset_w)
        self.w_preparation_t = np.array(w_preparation_t)
        self.w_rest_t = np.array(w_rest_t)
        self.mi_labels = np.array(mi_labels) if mi_labels is not None else \
            mi_labels
        self.mi_labels_info = mi_labels_info
        self.mi_result = np.array(mi_result) if mi_result is not None else \
            mi_result
        self.paradigm_info = paradigm_info

        # Optional attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
        return rec_dict

    @staticmethod
    def from_serializable_obj(dict_data):
        return MIData(**dict_data)


class MIDataset(components.Dataset):
    """This class inherits from medusa.data_structures.Dataset, increasing
    its functionality for datasets with data from MI experiments. It
    provides common ground for the rest of functions in the module.
    """

    def __init__(self, channel_set, fs=None, biosignal_att_key='eeg',
                 experiment_att_key='midata', experiment_mode=None,
                 track_attributes=None):
        """ Constructor

        Parameters
        ----------
        channel_set : meeg.EEGChannelSet
            EEG channel set. Only these channels will be kept in the dataset,
            the others will be discarded. Also, the signals will be rearranged,
            keeping the same channel order, avoiding errors in future stages of
            the signal processing pipeline
        fs : int, float or None
            Sample rate of the recordings. If there are recordings with
            different sample rates, the consistency of the dataset can be
            still assured using resampling
        biosignal_att_key : str
            Name of the attribute containing the target biosginal that will be
            used to extract the features. It has to be the same in all
            recordings (e.g., 'eeg', 'meg').
        experiment_att_key : str or None
            Name of the attribute containing the target experiment that will be
            used to extract the features. It has to be the same in all
            recordings (e.g., 'mi_left_right', 'rest_mi'). It is
            mandatory when a recording of the dataset contains more than 1
            experiment data
        experiment_mode : str {'train'|'test'|'guided test'|None}
            Mode of the experiment. If this dataset will be used to fit a model,
            set to train to avoid errors
        track_attributes: dict of dicts or None
            This parameter indicates custom attributes that must be tracked in
            feature extraction functions and how. The keys are the name of the
            attributes, whereas the values are dicts indicating the tracking
            mode {'concatenate'|'append'} and parent. Option concatenate is
            only available for attributes of type list or numpy arrays,
            forming a 1 dimensional array with the data from all recordings.
            Option append is used to save all kind of objects for each
            recording, forming a list whose length will be the number of
            recordings in the dataset. A set of default attributes is defined,
            so this parameter will be None in most cases. Example to track 2
            custom attributes (i.e., date and experiment_equipment):
                track_attributes = {
                    'date': {
                        'track_mode': 'append',
                        'parent': None
                    },
                    'experiment_equipment': {
                        'track_mode': 'append',
                        'parent': experiment_att_key
                    }
                }
        """
        # Check errors
        if experiment_mode is not None:
            if experiment_mode not in ('train', 'test', 'guided test'):
                raise ValueError('Parameter experiment_mode must be '
                                 '{train|test|guided test|None}')

        # Default track attributes
        # TODO: esto no va a funcionar con MIDataOld
        default_track_attributes = {
            'subject_id': {
                'track_mode': 'append',
                'parent': None
            },
            'onsets': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'w_trial_t': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'calibration_onset_w': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'w_preparation_t': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'w_rest_t': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'mi_labels_info': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'paradigm_info': {
                'track_mode': 'append',
                'parent': experiment_att_key
            }
        }

        if experiment_mode in ['train', 'guided test']:
            default_track_attributes_train = {
                'mi_labels': {
                    'track_mode': 'concatenate',
                    'parent': experiment_att_key
                }
            }
            default_track_attributes = {
                **default_track_attributes,
                **default_track_attributes_train
            }
        if experiment_mode in ['test', 'guided test']:
            default_track_attributes_test = {
                'mi_result': {
                    'track_mode': 'concatenate',
                    'parent': experiment_att_key
                }
            }
            default_track_attributes = {
                **default_track_attributes,
                **default_track_attributes_test
            }
        track_attributes = \
            default_track_attributes if track_attributes is None else \
                {**default_track_attributes, **track_attributes}

        # Class attributes
        self.channel_set = channel_set
        self.fs = fs
        self.biosignal_att_key = biosignal_att_key
        self.experiment_att_key = experiment_att_key
        self.experiment_mode = experiment_mode
        self.track_attributes = track_attributes

        # Consistency checker
        checker = self.__get_consistency_checker()
        super().__init__(consistency_checker=checker)

    def __get_consistency_checker(self):
        """Creates a standard consistency checker for MI datasets

        Returns
        -------
        checker : data_structures.ConsistencyChecker
            Standard consistency checker for MI feature extraction
        """
        # Create consistency checker
        checker = components.ConsistencyChecker()
        # Check that the biosignal exists
        checker.add_consistency_rule(
            rule='check-attribute',
            rule_params={'attribute': self.biosignal_att_key}
        )
        checker.add_consistency_rule(
            rule='check-attribute-type',
            rule_params={'attribute': self.biosignal_att_key,
                         'type': meeg.EEG}
        )
        # Check channels
        checker.add_consistency_rule(
            rule='check-values-in-attribute',
            rule_params={'attribute': 'channels',
                         'values': self.channel_set.channels},
            parent=self.biosignal_att_key + '.channel_set'
        )
        # Check sample rate
        if self.fs is not None:
            checker.add_consistency_rule(rule='check-attribute-value',
                                         rule_params={'attribute': 'fs',
                                                      'value': self.fs},
                                         parent=self.biosignal_att_key)
        else:
            warnings.warn('Parameter fs is None. The consistency of the '
                          'dataset cannot be assured. Still, you can use '
                          'target_fs parameter for feature extraction '
                          'and everything should be fine.')

        # Check experiment
        checker.add_consistency_rule(
            rule='check-attribute',
            rule_params={'attribute': self.experiment_att_key}
        )
        checker.add_consistency_rule(
            rule='check-attribute-type',
            rule_params={'attribute': self.experiment_att_key,
                         'type': MIData}
        )

        # Check track_attributes
        if self.track_attributes is not None:
            for key, value in self.track_attributes.items():
                checker.add_consistency_rule(
                    rule='check-attribute',
                    rule_params={'attribute': key},
                    parent=value['parent']
                )
                if value['track_mode'] == 'concatenate':
                    checker.add_consistency_rule(
                        rule='check-attribute-type',
                        rule_params={'attribute': key,
                                     'type': [list, np.ndarray, dict]},
                        parent=value['parent']
                    )

        return checker

    def custom_operations_on_recordings(self, recording):
        # Select channels
        eeg = getattr(recording, self.biosignal_att_key)
        eeg.change_channel_set(self.channel_set)
        return recording


class StandardPreprocessing(components.ProcessingMethod):
    """Just the common preprocessing applied in MI-based BCI. Simple,
    quick and effective: frequency IIR filter followed by common average
    reference (CAR) spatial filter.
    """

    def __init__(self, order=5, cutoff=[0.05, 63], btype='bandpass',
                 temp_filt_method='sosfiltfilt'):
        super().__init__(fit_transform_signal=['signal'],
                         fit_transform_dataset=['dataset'])
        # Parameters
        self.order = order
        self.cutoff = cutoff
        self.btype = btype
        self.temp_filt_method = temp_filt_method

        # Variables that
        self.iir_filter = None

    def fit(self, fs, n_cha=None):
        """Fits the IIR filter.

        Parameters
        ----------
        fs: float
            Sample rate of the signal.
        n_cha: int
            Number of channels. Used to compute the initial conditions of the
            frequency filter. Only required with sosfilt filtering method
            (online filtering)
        """
        self.iir_filter = IIRFilter(order=self.order,
                                    cutoff=self.cutoff,
                                    btype=self.btype,
                                    filt_method=self.temp_filt_method)
        self.iir_filter.fit(fs, n_cha=n_cha)

    def transform_signal(self, signal):
        """Transforms an EEG signal applying IIR filtering and CAR sequentially

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        """
        signal = self.iir_filter.transform(signal)
        signal = car(signal)
        return signal

    def fit_transform_signal(self, signal, fs):
        """Fits the IIR filter and transforms an EEG signal applying IIR
        filtering and CAR sequentially

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        fs: float
            Sample rate of the signal.
        """
        self.iir_filter = IIRFilter(order=self.order,
                                    cutoff=self.cutoff,
                                    btype=self.btype,
                                    filt_method=self.temp_filt_method)
        signal = self.iir_filter.fit_transform(signal, fs)
        signal = car(signal)
        return signal

    def fit_transform_dataset(self, dataset, show_progress_bar=True):
        """Fits the IIR filter and transforms an EEG signal applying the
        filter and CAR sequentially. Each recording is preprocessed
        independently, taking into account possible differences in sample rate.

        Parameters
        ----------
        dataset: MIDataset
            MIDataset with the recordings to be preprocessed.
        show_progress_bar: bool
            Show progress bar
        """
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Preprocessing')
        for rec in dataset.recordings:
            eeg = getattr(rec, dataset.biosignal_att_key)
            eeg.signal = self.fit_transform_signal(eeg.signal, eeg.fs)
            setattr(rec, dataset.biosignal_att_key, eeg)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return dataset


class StandardFeatureExtraction(components.ProcessingMethod):
    """Standard feature extraction method for MI-based spellers. Basically,
    it gets the raw epoch for each MI event.
    """

    def __init__(self, w_epoch_t=(0, 3000), target_fs=128,
                 baseline_mode='trial',
                 w_baseline_t=(-1500, -500), norm='z',
                 concatenate_channels=False, safe_copy=True):
        """Class constructor

        w_epoch_t : list
            Temporal window in ms for each epoch relative to the event onset
            (e.g., [0, 3000])
        target_fs : float of None
            Target sample rate of each epoch. If None, all the recordings must
            have the same sample rate, so it is strongly recommended to set this
            parameter to a suitable value to avoid problems and save time
        baseline_mode : {'run', 'trial', None}
            If "run" selected it will use the w_baseline_t of the beginning of
            each run.
            If "trial" selected it will use the w_baseline_t previous to each
            epoch extracted.
            If None it will not perform the baseline extraction.
        w_baseline_t : list, np.ndarray or "auto"
            Temporal window in ms to be used for baseline normalization.
            If baseline_mode = "run" it will use the temporal window since start
            of trial.
            If baseline_mode = "trial", for each epoch it will use the time
            relative to the event onset (e.g., [-1500, -500]).
            If "auto" selected:
                If baseline_mode = "run" will use the time between the
                MedusaData.experiment.start and MedusaData.experiment.calibration
                information stored in the run.
                If baseline_mode = "trial" will use the time between
                [-1500,-500]ms takes the baseline from -1500ms to -500ms before
                each onset (0 ms represents the onset).
                If baseline_mode = None it will not be used.
        norm : str {'z'|'dc'}
            Type of baseline normalization. Set to 'z' for Z-score normalization
            or 'dc' for DC normalization
        concatenate_channels : bool
            This parameter controls the shape of the feature array. If True, all
            channels will be concatenated, returning an array of shape [n_events
            x (samples x channels)]. If false, the array will have shape
            [n_events x samples x channels]
        safe_copy : bool
            Makes a safe copy of the signal to avoid changing the original
            samples due to references
        """
        super().__init__(transform_signal=['x'],
                         transform_dataset=['x', 'x_info'])
        self.w_epoch_t = w_epoch_t
        self.target_fs = target_fs
        self.baseline_mode = baseline_mode
        self.w_baseline_t = w_baseline_t
        self.norm = norm
        self.concatenate_channels = concatenate_channels
        self.safe_copy = safe_copy

    def transform_signal(self, times, signal, fs, onsets,
                         w_epoch_t = None):
        """Function to extract MI features from raw signal. It returns a 3D
        feature array with shape [n_events x n_samples x n_channels]. This
        function does not track any other attributes. Use for online processing
        and custom higher level functions.

        Parameters
        ----------
        times : list or numpy.ndarray
                1D numpy array [n_samples]. Timestamps of each sample. If they
                are not available, generate them artificially. Nevertheless,
                all signals and events must have the same temporal origin
        signal : list or numpy.ndarray
            2D numpy array [n_samples x n_channels]. EEG samples (the units
            should be defined using kwargs)
        fs : int or float
            Sample rate of the recording.
        onsets : list or numpy.ndarray [n_events x 1]
                Timestamp of each event

        Returns
        -------
        features : np.ndarray [n_events x n_samples x n_channels]
            Feature array with the epochs of signal
        """
        if w_epoch_t is None:
            w_epoch_t = self.w_epoch_t
        # Avoid changes in the original signal (this may not be necessary)
        if self.safe_copy:
            signal = signal.copy()
        # Baseline options
        norm = self.norm
        assert self.baseline_mode in ('run', 'trial', None), \
            ValueError('Parameter baseline_mode must be {"run", "trial", None}')
        if self.baseline_mode is None:
            assert self.w_baseline_t is None, 'If baseline_mode is None, ' \
                                              'parameter w_baseline_t must be ' \
                                              'None'
            w_baseline_trial_t = None
        elif self.baseline_mode == 'trial':
            if self.w_baseline_t == 'auto':
                w_baseline_trial_t = w_epoch_t
            else:
                w_baseline_trial_t = self.w_baseline_t
        elif self.baseline_mode == 'run':
            norm = None
            w_baseline_trial_t = None

        # Extract features
        features = get_epochs_of_events(timestamps=times, signal=signal,
                                        onsets=onsets, fs=fs,
                                        w_epoch_t=w_epoch_t,
                                        w_baseline_t=w_baseline_trial_t,
                                        norm=norm)

        if self.baseline_mode == "run":
            if self.w_baseline_t == 'auto':
                norm_epoch_t = [0, 5000]
            else:
                norm_epoch_t = self.w_baseline_t
            norm_epoch_s = np.array(norm_epoch_t * fs / 1000, dtype=int)
            norm_epoch = np.expand_dims(signal[norm_epoch_s], axis=0)
            features = normalize_epochs(features, norm_epochs=norm_epoch,
                                        norm=self.norm)
        # Resample each epoch to the target frequency
        if self.target_fs is not None:
            if self.target_fs > fs:
                raise warnings.warn('Target fs is greater than data fs')
            features = resample_epochs(features,
                                       self.w_epoch_t,
                                       self.target_fs)
        # Reshape epochs and concatenate the channels
        if self.concatenate_channels:
            features = np.squeeze(features.reshape((features.shape[0],
                                                    features.shape[1] *
                                                    features.shape[2], 1)))
        return features

    def transform_dataset(self, dataset, show_progress_bar=True,
                          continuous=False):
        #TODO: Review with modifications of Eduardo in erp_spellers (SERGIO)
        """High level function to easily extract features from EEG recordings
        and save useful info for later processing. Nevertheless, the provided
        functionality has several limitations and it will not be suitable for
        all cases and processing pipelines. If it does not fit your needs,
        create a custom function iterating the recordings and using
        extract_erp_features, a much more low-level and general function. This
        function does not apply any preprocessing to the signals, this must
        be done before

        Parameters
        ----------
        dataset: MIDataset
            List of data_structures.Recordings or data_structures.Dataset. If this
            parameter is a list of recordings, the consistency of the dataset will
            be checked. Otherwise, if the parameter is a dataset, this function
            assumes that the consistency is already checked
        show_progress_bar: bool
            Show progress bar

        Returns
        -------
        features : numpy.ndarray
            Array with the biosignal samples arranged in epochs
        track_info : dict
            Dictionary with tracked information across all recordings

        """
        # Avoid changes in the original recordings (this may not be necessary)
        if self.safe_copy:
            dataset = copy.deepcopy(dataset)
        # Avoid consistency problems
        if dataset.fs is None and self.target_fs is None:
            raise ValueError('The consistency of the features is not assured '
                             'since dataset.fs and target_fs are both None. '
                             'Specify one of these parameters')

        # Additional track attributes
        track_attributes = dataset.track_attributes
        # track_attributes['run_idx'] = {
        #     'track_mode': 'concatenate',
        #     'parent': dataset.experiment_att_key
        # }

        # Initialization
        features = None
        track_info = dict()
        for key, value in track_attributes.items():
            if value['track_mode'] == 'append':
                track_info[key] = list()
            elif value['track_mode'] == 'concatenate':
                track_info[key] = None
            else:
                raise ValueError('Unknown track mode')

        # Init progress bar
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Extracting features')

        # Compute features
        for run_counter, rec in enumerate(dataset.recordings):
            # Extract recording experiment and biosignal
            rec_exp = getattr(rec, dataset.experiment_att_key)
            rec_sig = getattr(rec, dataset.biosignal_att_key)

            # Get features
            # Get features
            list_w_epoch_t = list()
            if not continuous:
                list_w_epoch_t.append(self.w_epoch_t)
            if continuous:
                start = self.w_epoch_t - np.diff(self.w_epoch_t) / 2
                stop = rec_exp.w_trial_t[1] - self.w_epoch_t[1]
                steps_0 = np.arange(start=start[0], stop=stop + 1, step=250,
                                    dtype=int)
                steps_1 = np.arange(start=start[1], stop=rec_exp.w_trial_t[
                                                             1] + 1, step=250,
                                    dtype=int)
                for i in range(len(steps_0)):
                    list_w_epoch_t.append([steps_0[i], steps_1[i]])

            for w_epoch_t in list_w_epoch_t:

                rec_feat = self.transform_signal(
                    times=rec_sig.times,
                    signal=rec_sig.signal,
                    fs=rec_sig.fs,
                    onsets=rec_exp.onsets,
                    w_epoch_t=w_epoch_t
                    # ,
                    # calibration_t=rec_exp.calibration_t
                )

                features = np.concatenate((features, rec_feat), axis=0) \
                    if features is not None else rec_feat

                # Special attributes that need tracking across runs to assure the
                # consistency of the dataset
                # rec_exp.run_idx = run_counter * np.ones_like(rec_exp.trial_idx)

                # Track experiment info
                for key, value in track_attributes.items():
                    if value['parent'] is None:
                        parent = rec
                    else:
                        parent = rec
                        for p in value['parent'].split('.'):
                            parent = getattr(parent, p)
                    att = getattr(parent, key)
                    if value['track_mode'] == 'append':
                        track_info[key].append(att)
                    elif value['track_mode'] == 'concatenate':
                        track_info[key] = np.concatenate(
                            (track_info[key], att), axis=0
                        ) if track_info[key] is not None else att
                    else:
                        raise ValueError('Unknown track mode')

            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()

        return features, track_info


class CSPFeatureExtraction(StandardFeatureExtraction):
    """Common Spatial Patterns (CSP) feature extraction method for MI-based
    spellers.

    Processing pipeline:
    - Use of StandardFeatureExtraction to get the raw epoch of each MI event.
    - Extract CSP features of those MI events.
    """

    def __init__(self, n_filters=4, w_epoch_t=(500, 4000), target_fs=None,
                 baseline_mode='trial',
                 w_baseline_t=(-1500, -500), norm='z',
                 concatenate_channels=False, safe_copy=True, **kwargs):
        """Class constructor

        n_filter : int or None
            Number of most discriminant CSP filters to decompose the signal
            into (must be less or equal to the number of channels in your
            signal).
            If int it will return that number of filters.
            If None it will return the all calculated filters.
        w_epoch_t : list
            Temporal window in ms for each epoch relative to the event onset
            (e.g., [0, 3000])
        target_fs : float of None
            Target sample rate of each epoch. If None, all the recordings must
            have the same sample rate, so it is strongly recommended to set this
            parameter to a suitable value to avoid problems and save time
        baseline_mode : {'run', 'trial', None}
            If "run" selected it will use the w_baseline_t of the beginning of
            each run.
            If "trial" selected it will use the w_baseline_t previous to each
            epoch extracted.
            If None it will not perform the baseline extraction.
        w_baseline_t : list, np.ndarray or "auto"
            Temporal window in ms to be used for baseline normalization.
            If baseline_mode = "run" it will use the temporal window since start
            of trial.
            If baseline_mode = "trial", for each epoch it will use the time
            relative to the event onset (e.g., [0, 3000]).
            If "auto" selected:
                If baseline_mode = "run" will use the time between the
                MedusaData.experiment.start and MedusaData.experiment.calibration
                information stored in the run.
                If baseline_mode = "trial" will use the time between
                [-1500,-500]ms takes the baseline from -1500ms to -500ms before
                each onset (0 ms represents the onset).
                If baseline_mode = None it will not be used.
        norm : str {'z'|'dc'}
            Type of baseline normalization. Set to 'z' for Z-score normalization
            or 'dc' for DC normalization
        concatenate_channels : bool
            This parameter controls the shape of the feature array. If True, all
            channels will be concatenated, returning an array of shape [n_events
            x (samples x channels)]. If false, the array will have shape
            [n_events x samples x channels]
        safe_copy : bool
            Makes a safe copy of the signal to avoid changing the original
            samples due to references

        After using the function self.fit(), the attributes are computed:

        Attributes
        ----------
        CSP : CSP class with attributes filters, eigenvalues, patterns and
            methods fit and project.

            filters : {(…, M, M) numpy.ndarray, (…, M, M) matrix}
                Mixing matrix (spatial filters are stored in columns).
            eigenvalues : (…, M) numpy.ndarray
                Eigenvalues of w.
            patterns : numpy.ndarray
                De-mixing matrix (activation patterns are stored in columns).
        """
        super().__init__(w_epoch_t=w_epoch_t, target_fs=target_fs,
                         baseline_mode=baseline_mode,
                         w_baseline_t=w_baseline_t, norm=norm,
                         concatenate_channels=concatenate_channels,
                         safe_copy=safe_copy)
        self.CSP = CSP(n_filters=n_filters)

    def fit(self, X, y):
        """Train Common Spatial Patterns (CSP) filters

        Train Common Spatial Patterns (CSP) filters with support to >2
        classes based on support multiclass CSP by means of approximate
        joint diagonalization. In this case, the spatial filter selection
        is achieved according to [1].

        Adapted from http://github.com/alexandrebarachant/pyRiemann

        Parameters
        ----------
        X : numpy.ndarray, [n_trials, samples, channels]
            Epoched data of shape (n_trials, samples, channels)

        y : numpy.ndarray, [n_trials,]
            Labels for epoched data of shape (n_trials,)

        References
        ----------
        [1] Grosse-Wentrup, Moritz, and Martin Buss. "Multiclass common
        spatial patterns and information theoretic feature extraction."
        Biomedical Engineering, IEEE Transactions on 55, no. 8 (2008):
        1991-2000.
        """
        self.CSP.fit(X=X, y=y)

    def project(self, times, signal, fs, onsets):
        """Function to extract MI features from raw signal. It returns a 3D
        feature array with shape [n_events x n_samples x n_channels]. This
        function does not track any other attributes. Use for online processing
        and custom higher level functions.

        Parameters
        ----------
        times : list or numpy.ndarray
                1D numpy array [n_samples]. Timestamps of each sample. If they
                are not available, generate them artificially. Nevertheless,
                all signals and events must have the same temporal origin
        signal : list or numpy.ndarray
            2D numpy array [n_samples x n_channels]. EEG samples (the units
            should be defined using kwargs)
        fs : int or float
            Sample rate of the recording.
        onsets : list or numpy.ndarray [n_events x 1]
                Timestamp of each event

        Returns
        -------
        features : numpy.ndarray [n_events x n_samples x n_channels]
            Feature array with the epochs of signal projected in the CSP space
        """
        features = super().transform_signal(times, signal, fs, onsets)

        features = self.CSP.project(X=features)
        return features

    def extract_log_var_features(self, X):
        """This method computes the standard motor imagery log-variance features
        given the CSP .

        Parameters
        ----------
        X : numpy.ndarray, [n_trials, samples, channels]
            Epoched data of shape (n_trials, samples, channels)
        Returns
        -------
        features : numpy.ndarray [n_events x n_channels]
            Feature array with the epochs of signal projected in the CSP space
        """
        features = self.CSP.project(X=X)
        # Obtain log-variance features given the CSP features
        features = np.log(np.var(features, axis=-1))
        return features

    def transform_dataset(self, dataset, show_progress_bar=True):
        features, track_info = super().transform_dataset(
            dataset=dataset, show_progress_bar=show_progress_bar)
        # Fit CSP filter
        self.fit(X=features, y=track_info['mi_labels'])
        # Project features into CSP space and obtain log-variance features given
        # the CSP features
        features = self.extract_log_var_features(X=features)

        return features, track_info


class MIModel(components.Algorithm):
    """Skeleton class for MI-based BCIs models. This class inherits from
    components.Algorithm. Therefore, it can be used to create standalone
    algorithms that can be used in compatible apps from medusa-platform
    for online experiments. See components.Algorithm to know more about this
    functionality.

    Related tutorials:

        - Overview of mi_paradigms module [LINK]
        - Create standalone models for MI-based BCIs compatible with
            Medusa platform [LINK]
    """

    def __init__(self):
        """Class constructor
        """
        super().__init__(fit_dataset=['mi_target', 'mi_result', 'accuracy'],
                         predict=['mi_result'])
        # Settings
        self.settings = None
        self.channel_set = None
        self.configure()
        # Configuration
        self.is_configured = False
        self.is_built = False
        self.is_fit = False

    @abstractmethod
    def configure(self, **kwargs):
        """This function must be used to configure the model before calling
        build method. Class attribute settings attribute must be set with a dict
        """
        # Update state
        self.is_configured = True
        self.is_built = False
        self.is_fit = False

    @abstractmethod
    def build(self, *args, **kwargs):
        """This function builds the model, adding all the processing methods
        to the pipeline. It must be called after configure.
        """
        # Check errors
        if not self.is_configured:
            raise ValueError('Function configure must be called first!')
        # Update state
        self.is_built = True
        self.is_fit = False

    def fit_dataset(self, dataset, **kwargs):
        """Function that receives an MIDataset and uses its data to
        fit the model. By default, executes pipeline 'fit_dataset'. Override
        method for other behaviour.

        Parameters
        ----------
        dataset: MIDataset
            Dataset with recordings from an MI-based BCI experiment
        kwargs: key-value arguments
            Optional parameters depending on the specific implementation of
            the model

        Returns
        -------
        fit_results: dict
            Dict with the information of the fit process. For command
            decoding models, at least it has to contain keys
            mi_target, mi_result and accuracy, which contain the target MI,
            the decoded MI and the decoding accuracy in the analysis.
        """
        # Check errors
        if not self.is_built:
            raise ValueError('Function build must be called first!')
        # Execute pipeline
        output = self.exec_pipeline('fit_dataset', dataset=dataset)
        # Set channels
        self.channel_set = dataset.channel_set
        # Update state
        self.is_fit = True
        return output

    def predict(self, times, signal, fs, l_cha, x_info, **kwargs):
        """Function that receives EEG signal and experiment info from an
        MI-based trial to decode the user's intentions. Used in online
        experiments. By default, executes pipeline 'predict'. Override method
        for other behaviour.

        Parameters
        ---------
        times: list or numpy.ndarray
            Timestamps of the EEG samples
        signal: list or numpy.ndarray
            EEG samples with shape [n_samples x n_channels]
        fs: float
            Sample rate of the EEG signal
        l_cha: list
            List of channel labels
        x_info: dict
            Dict with the needed experiment info to decode the commands. It
            has to contain keys: mode, onsets, w_trial_t.
            See MIData to know how are defined these variables.
        kwargs: key-value arguments
            Optional parameters depending on the specific implementation of
            the model
        """
        # Check errors
        if not self.is_fit:
            raise ValueError('Function fit_dataset must be called first!')
        # Check channels
        self.channel_set.check_channels_labels(l_cha, strict=True)
        # Execute predict pipeline
        return self.exec_pipeline('predict', times=times, signal=signal,
                                  fs=fs, x_info=x_info, **kwargs)


class MIModelCSP(MIModel):
    """Decoding model for MI-based BCI applications based on
    Common Spatial Patterns (CSP).

    Dataset features:

    - Sample rate of the signals > 60 Hz. The model can handle recordings
        with different sample rates.
    - Recommended channels: ['C3', 'C4'].

    Processing pipeline:
    - Preprocessing (medusa.bci.erp_spellers.StandardPreprocessing):

        - IIR Filter (order=5, cutoff=(8, 30) Hz: unlike FIR filters, IIR
            filters are quick and can be applied in small signal chunks. Thus,
            they are the preferred method for frequency filter in online
            systems.
        - Common average reference (CAR): widely used spatial filter that
            increases the signal-to-noise ratio of the MI control signals.
    - Feature extraction (medusa.bci.mi_paradigms.CSPFeatureExtraction):

        - Epochs (window=(500, 4000) ms, resampling to 60 HZ): the epochs of
            signal are extracted for each stimulation. Baseline normalization
            is also applied, taking the window (-1500, -500) ms relative to the
            stimulus onset.
        - CSP projection: Epochs are then projected according to a CSP filter
            previously trained.

    - Feature classification (
    sklearn.discriminant_analysis.LinearDiscriminantAnalysis)

        - Regularized linear discriminant analysis (rLDA): we use the sklearn
            implementation, with eigen solver and auto shrinkage paramers. See
            reference in sklearn doc.
    """

    def __init__(self):
        super().__init__()

    def configure(self, p_filt_cutoff=(8, 30), f_w_epoch_t=(500, 4000),
                  f_target_fs=60, **kwargs):
        self.settings = {
            'p_filt_cutoff': p_filt_cutoff,
            'f_w_epoch_t': f_w_epoch_t,
            'f_target_fs': f_target_fs
        }
        self.settings = dict(self.settings, **kwargs)
        # Update state
        self.is_configured = True
        self.is_built = False
        self.is_fit = False

    def build(self):
        # Check errors
        if not self.is_configured:
            raise ValueError('Function configure must be called first!')
        # Preprocessing (default: bandpass IIR filter [8, 30] Hz + CAR)
        self.add_method('prep_method', StandardPreprocessing(
            cutoff=self.settings['p_filt_cutoff']
        ))
        # Feature extraction (default: epochs [500, 4000] ms + resampling to 80
        # Hz)
        self.add_method('ext_method', CSPFeatureExtraction(**self.settings))
        # Feature classification (rLDA)
        clf = components.ProcessingClassWrapper(
            LinearDiscriminantAnalysis(solver='eigen', shrinkage='auto'),
            fit=[], predict_proba=['y_pred']
        )
        self.add_method('clf_method', clf)
        # Update state
        self.is_built = True
        self.is_fit = False

    def fit_dataset(self, dataset, **kwargs):
        # Check errors
        if not self.is_built:
            raise ValueError('Function build must be called first!')
        # Preprocessing
        dataset = self.get_inst('prep_method').fit_transform_dataset(dataset)
        # Extract CSP features
        x, x_info = self.get_inst('ext_method').transform_dataset(dataset)

        # Classification
        self.get_inst('clf_method').fit(x, x_info['mi_labels'])
        y_prob = self.get_inst('clf_method').predict_proba(x)
        y_pred = self.get_inst('clf_method').predict(x)

        # Accuracy
        accuracy = np.sum((y_pred == x_info['mi_labels'])) / len(y_pred)
        clf_report = classification_report(x_info['mi_labels'], y_pred,
                                           output_dict=True)
        assessment = {
            'x': x,
            'x_info': x_info,
            'y_prob': y_prob,
            'y_pred': y_pred,
            'accuracy': accuracy,
            'report': clf_report
        }
        # Save info
        self.channel_set = dataset.channel_set
        # Update state
        self.is_fit = True
        return assessment

    def predict(self, times, signal, fs, channel_set, x_info, **kwargs):
        # Check errors
        if not self.is_fit:
            raise ValueError('Function fit_dataset must be called first!')
        # Check channel set
        if self.channel_set.channels != channel_set.channels:
            warnings.warn('The channel set is not the same that was used to '
                          'fit the model. Be careful!')
        # Preprocessing
        signal = self.get_inst('prep_method').fit_transform_signal(signal, fs)

        # Extract features
        x = self.get_inst('ext_method').transform_signal(times, signal, fs,
                                                         x_info['onsets'])
        x = self.get_inst('ext_method').extract_log_var_features(x)
        # Classification
        y_prob = self.get_inst('clf_method').predict_proba(x)
        y_pred = self.get_inst('clf_method').predict(x)

        # Decoding
        accuracy = None
        clf_report = None
        if x_info['mi_labels'] is not None:
            accuracy = np.sum((y_pred == x_info['mi_labels'])) / len(y_pred)
            clf_report = classification_report(x_info['mi_labels'], y_pred,
                                               output_dict=True)
        decoding = {
            'x': x,
            'x_info': x_info,
            'y_prob': y_prob,
            'y_pred': y_pred,
            'accuracy': accuracy,
            'report': clf_report
        }
        return decoding


class MIModelEEGSym(MIModel):
    """Decoding model for MI-based BCI applications based on EEGSym [1], a deep
    convolutional neural network developed for inter-subjects MI classification.

    Dataset features:

    - Sample rate of the signals > 128 Hz. The model can handle recordings
        with different sample rates.
    - Recommended channels: ['F7', 'C3', 'Po3', 'Cz', 'Pz', 'F8', 'C4', 'Po4'].

    Processing pipeline:

    - Preprocessing:

        - IIR Filter (order=4, lowpass=49 Hz: unlike FIR filters, IIR filters
            are quick and can be applied in small signal chunks. Thus,
            they are the preferred method for frequency filter in online systems
        - Common average reference (CAR): widely used spatial filter that
            increases the signal-to-noise ratio.
    - Feature extraction:

        - Epochs (window=(0, 2000) ms, resampling to 128 HZ): the epochs of
            signal are extracted after each onset. Baseline normalization
            is also applied, taking the same epoch window.

    - Feature classification

        - EEGSym: convolutional neural network [1].

    References
    ----------
    [1] Pérez-Velasco, S., Santamaría-Vázquez, E., Martínez-Cagigal, V.,
    Marcos-Mateo, D., & Hornero, R. (2020). EEGSym: Overcoming Intersubject
    Variability in Motor Imagery Based BCIs with Deep Learning. ?.
    """
    def __init__(self):
        super().__init__()

    def configure(self, cnn_n_cha=8, ch_lateral=3, fine_tuning=False,
                  shuffle_before_fit=True, validation_split=0.4,
                  init_weights_path=None, gpu_acceleration=False,
                  augmentation=False):
        self.settings = {
            'cnn_n_cha': cnn_n_cha,
            'ch_lateral': ch_lateral,
            'fine_tuning': fine_tuning,
            'augmentation': augmentation,
            'shuffle_before_fit': shuffle_before_fit,
            'validation_split': validation_split,
            'init_weights_path': init_weights_path,
            'gpu_acceleration': gpu_acceleration
        }
        # Update state
        self.is_configured = True
        self.is_built = False
        self.is_fit = False

    def build(self):
        # Check errors
        if not self.is_configured:
            raise ValueError('Function configure must be called first!')
        # Only import deep learning models if necessary
        from medusa.deep_learning_models import EEGSym
        # Preprocessing (bandpass IIR filter [0.5, 45] Hz + CAR)
        self.add_method('prep_method',
                        StandardPreprocessing(cutoff=49, btype='lowpass'))
        # Feature extraction (epochs [0, 2000] ms + resampling to 128 Hz)
        self.add_method('ext_method', StandardFeatureExtraction(w_epoch_t=(
            0, 2000), target_fs=128, w_baseline_t=(0, 2000),))
        # Feature classification
        clf = EEGSym(
            input_time=2000,
            fs=128,
            n_cha=self.settings['cnn_n_cha'],
            ch_lateral=self.settings['ch_lateral'],
            filters_per_branch=24,
            scales_time=(125, 250, 500),
            dropout_rate=0.4,
            activation='elu', n_classes=2,
            learning_rate=0.0001,
            gpu_acceleration=self.settings['gpu_acceleration'])
        self.is_fit = False
        if self.settings['init_weights_path'] is not None:
            clf.load_weights(self.settings['init_weights_path'])
            self.channel_set = meeg.EEGChannelSet()
            standard_lcha = ['F7', 'C3', 'Po3', 'Cz', 'Pz', 'F8', 'C4', 'Po4']
            self.channel_set.set_standard_montage(standard_lcha)
            self.is_fit = True
        self.add_method('clf_method', clf)
        # Update state
        self.is_built = True
        # self.is_fit = False

    def fit_dataset(self, dataset, continuous=False, **kwargs):
        # Check errors
        if not self.is_built:
            raise ValueError('Function build must be called first!')
        # Preprocessing
        dataset = self.get_inst('prep_method').fit_transform_dataset(dataset)
        # Extract features
        x, x_info = self.get_inst('ext_method').transform_dataset(dataset,
                                                                  continuous=continuous)
        # Put channels in symmetric order
        x = self.get_inst('clf_method').symmetric_channels(x,
                                                           dataset.channel_set.l_cha)

        # Classification
        self.get_inst('clf_method').fit(x, x_info['mi_labels'],
                                        fine_tuning=self.settings['fine_tuning'],
                                        shuffle_before_fit=self.settings['shuffle_before_fit'],
                                        validation_split=self.settings['validation_split'],
                                        augmentation=self.settings['augmentation'],
                                        **kwargs)
        y_prob = self.get_inst('clf_method').predict_proba(x)
        y_pred = y_prob.argmax(axis=-1)

        # Accuracy
        accuracy = np.sum((y_pred == x_info['mi_labels'])) / len(y_pred)
        clf_report = classification_report(x_info['mi_labels'], y_pred,
                                           output_dict=True)
        assessment = {
            'x': x,
            'x_info': x_info,
            'y_pred': y_pred,
            'y_prob': y_prob,
            'accuracy': accuracy,
            'report': clf_report
        }
        # Save info
        self.channel_set = dataset.channel_set
        # Update state
        self.is_fit = True
        return assessment

    def predict(self, times, signal, fs, channel_set, x_info, **kwargs):
        # Check errors
        if not self.is_fit:
            raise ValueError('Function fit_dataset must be called first!')
        # Check channel set
        if self.channel_set.channels != channel_set.channels:
            warnings.warn('The channel set is not the same that was used to '
                          'fit the model. Be careful!')
        # Preprocessing
        signal = self.get_inst('prep_method').fit_transform_signal(signal, fs)

        # Extract features
        x = self.get_inst('ext_method').transform_signal(times, signal, fs,
                                                         x_info['onsets'])
        # Put channels in symmetric order
        x = self.get_inst('clf_method').symmetric_channels(x,
                                                           self.channel_set.l_cha)
        # Classification
        y_prob = self.get_inst('clf_method').predict_proba(x)
        y_pred = y_prob.argmax(axis=-1)

        # Decoding
        accuracy = None
        clf_report = None
        if x_info['mi_labels'] is not None:
            accuracy = np.sum((y_pred == x_info['mi_labels'])) / len(y_pred)
            clf_report = classification_report(x_info['mi_labels'], y_pred,
                                               output_dict=True)
        decoding = {
            'x': x,
            'x_info': x_info,
            'y_pred': y_pred,
            'y_prob': y_prob,
            'accuracy': accuracy,
            'report': clf_report
        }
        return decoding


"""
Created on Tue Sep 27 10:45:15 2022

@author: Víctor Martínez-Cagigal
"""
import matplotlib.pyplot as plt
import numpy as np


def shaded_plot(data, error='std', times=None, axes=None):
    """ This function plots the mean and the shaded error of a time series.

    Parameters
    ----------------
    data: 2D numpy.ndarray
        Data being plotted, with dimensions [observations x signal]
    error: basestring
        Type of error being plotted (mean+error, mean-error), which can be:
        - 'std':    standard deviation
        - 'sem':    standard error mean
        - 'var':    variance
        - 'c95':    95% confidence interval (assuming normality)
    times: numpy.ndarray
        Timestamps of the time series.
    axes: matplotlib.pyplot.axes
        If a matplotlib axes are specified, the plot is displayed inside it.
        Otherwise, the plot will generate a new figure.

    Returns
    ----------------
    axes: matplotlib.pyplot.axes
    """
    # Error detection
    must_show = False
    if axes is None:
        axes = plt.subplot(111)
        must_show = True
    if not isinstance(data, (np.ndarray, np.generic)):
        data = np.array(data)
    if len(data.shape) != 2:
        raise ValueError("Parameter 'data' must have 2 dimensions!")
    if times is None:
        times = np.linspace(0, data.shape[1], data.shape[1])

    # Mean and std
    mean_ = np.mean(data, axis=0)
    std_ = np.std(data, axis=0)

    # Error
    if error == 'std':
        error_ = std_
    elif error == 'sem':
        error_ = std_ / np.sqrt(data.shape[0])
    elif error == 'var':
        error_ = std_ ** 2
    elif error == 'c95':
        # assuming normality
        error_ = (std_ / np.sqrt(data.shape[0])) * 1.96
    else:
        raise ValueError("Error %s not recognized! Please, use std, sem, "
                         "var or c95" % error)

    # Plot
    axes.fill_between(times, mean_ + error_, mean_ - error_, alpha=0.3)
    axes.plot(times, mean_, linewidth=2)
    if must_show:
        plt.show()

    return axes



"""Created on Monday March 15 19:27:14 2021

This module provides

@author: Eduardo Santamaría-Vázquez
"""

import tensorflow as tf
import os
import warnings


class NoGPU(Exception):
    def __init__(self):
        super().__init__('No GPU available')


class NoCPU(Exception):
    def __init__(self):
        super().__init__('No CPU available')


class DeviceNotAvailable(Exception):
    def __init__(self, device):
        super().__init__('Device %s not available' % device)


def config_tensorflow(gpu_acceleration=True, device=None):
    """
    This function makes a very simple configuration of tensorflow, checking if
    Tensorflow >= 2.0.0 and GPU acceleration is available. For more advanced
    options, use the tensorflow native API.

    Parameters
    ----------
    gpu_acceleration: bool
        Enables GPU acceleration. The function will check if this feature is
        available and select the proper device.
    device: str, optional
        Device that will be selected. Use tf.config.list_logical_devices() to
        list the current available devices.
    """
    try:
        if int(tf.__version__.split('.')[0]) < 2:
            raise ImportError('Tensorflow >= 2.0.0 is required for GPU '
                              'acceleration')
        # Get available devices
        gpus = tf.config.list_logical_devices(device_type='GPU')
        cpus = tf.config.list_logical_devices(device_type='CPU')
        if gpu_acceleration:
            if len(gpus) > 0:
                os.environ["MEDUSA_TF_GPU_ACCELERATION"] = "1"
                if device is None:
                    os.environ["MEDUSA_TF_DEVICE"] = gpus[0].name
                else:
                    check = False
                    for gpu in gpus:
                        if gpu.name == device:
                            check = True
                    if check:
                        os.environ["MEDUSA_TF_DEVICE"] = device
                    else:
                        raise Exception('Device %s not available' % device)
            else:
                raise NoGPU()
        else:
            if len(cpus) > 0:
                os.environ["MEDUSA_TF_GPU_ACCELERATION"] = "0"
                if device is None:
                    os.environ["MEDUSA_TF_DEVICE"] = cpus[0].name
                else:
                    check = False
                    for cpu in cpus:
                        if cpu.name == device:
                            check = True
                    if check:
                        os.environ["MEDUSA_TF_DEVICE"] = device
                    else:
                        raise DeviceNotAvailable(device)
            else:
                raise NoCPU()

            warnings.warn('Some medusa modules take great advantage from GPU '
                          'acceleration to increase the performance of medusa. '
                          'You should consider execute your program in a device'
                          'with available GPU.')

        print('GPU acceleration: %s' % os.environ["MEDUSA_TF_GPU_ACCELERATION"])
        print('Selected device: %s' % os.environ["MEDUSA_TF_DEVICE"])

    except ModuleNotFoundError:
        raise ModuleNotFoundError('Tensorflow is not installed')


def check_tf_config(autoconfig=False):
    """Checks if tensorflow has been configured

     Parameters
    ----------
    autoconfig: bool
        If tensorflow has not been configured and autoconfig is True,
        tensorflow is configured automatically, trying GPU first.
    """
    check = True if os.environ.get("MEDUSA_TF_GPU_ACCELERATION") is not None \
        else False
    if not check and autoconfig:
        __auto_config_tensorflow()
        check = True if os.environ.get("MEDUSA_TF_GPU_ACCELERATION") is not \
                        None else False
    return check


def check_gpu_acceleration():
    """Checks if there is GPU acceleration available"""
    # Check configuration
    if not check_tf_config():
        raise Exception('Tensorflow has not been configured. Call function '
                        'medusa.gpu_acceleration.config_tensorflow')
    # Check GPU
    return True if os.environ.get("MEDUSA_TF_GPU_ACCELERATION") == '1' \
        else False


def get_tf_device_name():
    """Returns the tensorflow device used by medusa"""
    # Check configuration
    if not check_tf_config():
        raise Exception('Tensorflow has not been configured. Call function '
                        'medusa.gpu_acceleration.config_tensorflow')
    return os.environ["MEDUSA_TF_DEVICE"]


def __auto_config_tensorflow():
    """
    This function tries to configure tensorflow automatically. It is meant for
    internal use within medusa-core and avoid errors, but should not be used in
    custom scripts. Use config_tensorflow instead.
    """
    # Configure tensorflow integration automatically if it is not done yet
    try:
        config_tensorflow()
    except NoGPU as e:
        config_tensorflow(gpu_acceleration=False)
    warnings.warn('Tensorflow configured automatically. Function '
                  'tensorflow_integration.config_tensorflow should be'
                  ' called before for custom behaviour.')


import numpy as np
from scipy import signal as scipy_signal
import matplotlib.pyplot as plt
from medusa import components


class FIRFilter(components.ProcessingMethod):

    def __init__(self, order, cutoff, btype, width=None, window='hamming',
                 scale=True, filt_method='filtfilt', axis=0):
        """FIR filter designed using the implementation of scipy.signal.firwin.
        See the documentation of this function to find useful information about
        this class

        Parameters
        ----------
        order: int
            Length of the filter (number of coefficients, i.e. the filter order
            + 1). numtaps must be odd if a passband includes the Nyquist
            frequency.
        cutoff: float or 1-D array_like
            Cutoff frequency of filter (expressed in the same units as fs) OR an
            array of cutoff frequencies (that is, band edges). In the latter
            case, the frequencies in cutoff should be positive and monotonically
            increasing between 0 and fs/2. The values 0 and fs/2 must not be
            included in cutoff.
        btype: str {‘bandpass’|‘lowpass’|‘highpass’|‘bandstop’}
            Band type of the filter. It also controls the parameter pass_zero of
            them scipy.signal.firwin function
        width: float or None, optional
            If width is not None, then assume it is the approximate width of the
            transition region (expressed in the same units as fs) for use in
            Kaiser FIR filter design. In this case, the window argument is
            ignored.
        window: string or tuple of string and parameter values, optional
            Desired window to use. See scipy.signal.get_window for a list of
            windows and required parameters.
        scale: bool, optional
            Set to True to scale the coefficients so that the frequency response
            is exactly unity at a certain frequency. That frequency is either:

                - 0 (DC) if the first passband starts at 0 (i.e. pass_zero is
                    True)
                - fs/2 (the Nyquist frequency) if the first passband ends at
                    fs/2 (i.e the filter is a single band highpass filter);
                    center of first passband otherwise
        filt_method: str {'lfilter', 'filtfilt'}
            Filtering method. See scipy.signal.lfilter or scipy.signal.filtfilt
            for more information.
        axis: int
            The axis to which the filter is applied. By convention, signals
            in medusa are defined by [samples x channels], so axis is set to
            0 by default.
        """
        # Super call to specify the outputs of fit and apply functions
        super().__init__(fit=[], transform=['s'], fit_transform=['s'])

        # Variables
        self.btype = btype
        self.order = order
        self.cutoff = cutoff
        self.width = width
        self.window = window
        self.scale = scale
        self.filt_method = filt_method
        self.axis = axis

        # Parameters to fit
        self.fs = None
        self.a = None
        self.b = None

    def display(self):
        display_filter(self.b, self.a, self.fs)

    def fit(self, fs):
        self.fs = fs
        self.b = scipy_signal.firwin(numtaps=self.order,
                                     cutoff=self.cutoff,
                                     width=self.width,
                                     window=self.window,
                                     pass_zero=self.btype,
                                     scale=self.scale,
                                     fs=self.fs)
        self.a = [1.0]

    def transform(self, signal):
        if self.filt_method == 'filtfilt':
            s = scipy_signal.filtfilt(self.b, self.a, signal,
                                      axis=self.axis)
        elif self.filt_method == 'lfilter':
            s = scipy_signal.lfilter(self.b, self.a, signal,
                                     axis=self.axis)
        else:
            raise ValueError("Unsupported filtering method method!")
        return s

    def fit_transform(self, signal, fs):
        """Fits and applies the filter

        Parameters
        ----------
        signal: np.ndarray
            Signal to filter. By default, the expected shape is [samples x
            channels], but this order can be changed using axis parameter in
            constructor.
        fs: float
            The sampling frequency of the signal in Hz. Each frequency in
            cutoff must be between 0 and fs/2. Default is 2.
        """
        self.fit(fs)
        return self.transform(signal)


class IIRFilter(components.ProcessingMethod):

    def __init__(self, order, cutoff, btype, filt_method='sosfiltfilt', axis=0):
        """IIR Butterworth filter wrapper designed using implementation of
        scipy.signal.butter. See the documentation of this function to find
        useful information about this class.

        Parameters
        ----------
        order: int
            Length of the filter (number of coefficients, i.e. the filter order
            + 1). This parameter must be odd if a passband includes the
            Nyquist frequency.
        cutoff: float or 1-D array_like
            Cutoff frequency of filter (expressed in the same units as fs) OR an
            array of cutoff frequencies (that is, band edges). In the latter
            case, the frequencies in cutoff should be positive and monotonically
            increasing between 0 and fs/2. The values 0 and fs/2 must not be
            included in cutoff.
        btype: str {‘bandpass’|‘lowpass’|‘highpass’|‘bandstop’}
            Band type of the filter. It also controls the parameter pass_zero of
            them scipy.signal.firwin function
        filt_method: str {'sosfilt', 'sosfiltfilt'}
            Filtering method. See scipy.signal.sosfilt or
            scipy.signal.sosfiltfilt for more information. For real time
            fitlering, use sosfilt. For offline filtering, sosfiltfilt is the
            recommended filtering method.
        axis: int
            The axis to which the filter is applied. By convention, signals
            in medusa are defined by [samples x channels], so axis is set to
            0 by default.
        """
        # Super call to specify the outputs of fit and apply functions
        super().__init__(fit=[], transform=['s'], fit_transform=['s'])

        # Variables
        self.btype = btype
        self.order = order
        self.cutoff = cutoff
        self.filt_method = filt_method
        self.axis = axis

        # Parameters to fit
        self.fs = None
        self.sos = None
        self.zi = None

    def display(self):
        """Displays the filter. Function fit must be called first. This uses
        the function medusa.frequency_filtering.display_filter()
        """
        b, a = scipy_signal.sos2tf(self.sos)
        display_filter(b, a, self.fs)

    def fit(self, fs, n_cha=None):
        """Fits the filter

        Parameters
        ----------
        fs: float
            The sampling frequency of the signal in Hz. Each frequency in
            cutoff must be between 0 and fs/2. Default is 2.
        n_cha: int
            Number of channels. Used to compute the initial conditions of the
            filter. Only required with sosfilt filtering method (online
            filtering)
        """
        self.fs = fs
        self.sos = scipy_signal.butter(N=self.order,
                                       Wn=self.cutoff,
                                       btype=self.btype,
                                       analog=False,
                                       output='sos',
                                       fs=self.fs)
        if self.filt_method == 'sosfilt':
            if n_cha is None:
                raise ValueError('Specify the number of channels to compute '
                                 'the initial conditions of the filter')
            self.zi = scipy_signal.sosfilt_zi(self.sos)
            self.zi = np.repeat(self.zi[:, :, np.newaxis], n_cha, axis=2)

    def transform(self, signal):
        """Applies the filter to the signal

        Parameters
        ----------
        signal: np.ndarray
            Signal to filter. By default, the expected shape is [samples x
            channels], but this order can be changed using axis parameter in
            constructor.
        """
        if self.filt_method == 'sosfiltfilt':
            s = scipy_signal.sosfiltfilt(self.sos, signal, axis=self.axis)
        elif self.filt_method == 'sosfilt':
            s, zo = scipy_signal.sosfilt(self.sos, signal, axis=self.axis,
                                         zi=self.zi)
            self.zi = zo
        else:
            raise ValueError("Unsupported filtering method method!")
        return s

    def fit_transform(self, signal, fs):
        """Fits and applies the filter

        Parameters
        ----------
        signal: np.ndarray
            Signal to filter. By default, the expected shape is [samples x
            channels], but this order can be changed using axis parameter in
            constructor.
        fs: float
            The sampling frequency of the signal in Hz. Each frequency in
            cutoff must be between 0 and fs/2. Default is 2.
        """
        if self.axis == 0:
            n_cha = signal.shape[1]
        else:
            n_cha = signal.shape[0]
        self.fit(fs, n_cha)
        return self.transform(signal)


def display_filter(b, a, fs):
    """Displays the frequency response of a given filter.

    Parameters
    ----------
    b: np.ndarray
        Numerator of the filter
    a: np.ndarray
        Denominator of the filter
    fs: float
        Sampling frequency of the signal (in Hz)
    """
    # Frequency response
    w, h = scipy_signal.freqz(b, a)
    freq = w*fs/(2*np.pi)

    # Plot
    fig, ax = plt.subplots(2, 1, figsize=(8, 6))

    # Frequency response
    ax[0].plot(freq, 20 * np.log10(abs(h)), color='blue')
    ax[0].set_title("Frequency Response")
    ax[0].set_ylabel("Amplitude (dB)", color='blue')
    ax[0].set_xlim([0, fs/2])
    ax[0].set_ylim([-50, 1])
    ax[0].grid()

    # Phase response
    ax[1].plot(freq, np.unwrap(np.angle(h)) * 180 / np.pi, color='green')
    ax[1].set_ylabel("Angle (degrees)", color='green')
    ax[1].set_xlabel("Frequency (Hz)")
    ax[1].set_xlim([0, fs/2])
    ax[1].set_yticks([-90, -60, -30, 0, 30, 60, 90])
    ax[1].set_ylim([-90, 90])
    ax[1].grid()

    plt.show()








import numpy as np


def absolute_band_power(psd, fs, target_band):
    """This method computes the absolute band power of the signal in the given
    band using the power spectral density (PSD).

    Parameters
    ----------
    psd : numpy array
        PSD of the signal with shape [n_epochs, n_samples, n_channels]. Some
        of these dimensions may not exist in advance. In these case, create new
        axis using np.newaxis. E.g., non-epoched single-channel psd with
        shape [n_samples] can be passed to this function with psd[numpy.newaxis,
        ..., numpy.newaxis]. Afterwards, you may use numpy.squeeze to eliminate
        those axes.
    fs : int
        Sampling frequency of the signal
    target_band : numpy 2D array
        Frequency band where to calculate the power in Hz. E.g., [8, 13]

    Returns
    -------
    powers : numpy 2D array
        RP value for each band, epoch and channel. [n_bands, n_epochs,
        n_channels]
    """
    # To numpy arrays
    psd = np.array(psd)

    # Check errors
    if len(psd.shape) != 3:
        raise Exception('Parameter psd must have shape [n_epochs x n_samples x '
                        'n_channels]')

    # Calculate freqs array
    freqs = np.linspace(0, fs/2, psd.shape[1])

    # Compute power
    psd_target_samp = \
        np.logical_and(freqs >= target_band[0], freqs <= target_band[1])
    band_power = np.sum(psd[:, psd_target_samp, :], axis=1) * \
                 (fs / (2 * freqs.shape[0]))

    return band_power


def relative_band_power(psd, fs, target_band, baseline_band=None):
    """This method computes the relative band power of the signal in the given
    band using the power spectral density (PSD). Do not use this method on PSDs
    that are already normalized! In this case, use absolute_band_power function.

    Parameters
    ----------
    psd : numpy array
        PSD with shape [n_epochs, n_samples, n_channels]. Some of these
        dimensions may not exist in advance. In these case, create new axis
        using np.newaxis. E.g., non-epoched single-channel psd with shape
        [n_samples] can be passed to this function with psd[numpy.newaxis, ...,
        numpy.newaxis]. Afterwards, you may use numpy.squeeze to eliminate
        those axes.
    fs : int
        Sampling frequency of the signal
    target_band : numpy nd array
        Frequency band where to calculate the power in Hz. E.g., [8, 13]
    baseline_band: numpy nd array or None
        Frequency band where used as baseline in Hz. Leave to None to normalize
        by the whole spectrum, which is preferred in most cases

    Returns
    -------
    powers : numpy 2D array
        RP value for each band, epoch and channel. [n_bands, n_epochs,
        n_channels]
    """
    # To numpy arrays
    psd = np.array(psd)

    # Check errors
    if len(psd.shape) != 3:
        raise Exception('Parameter psd must have shape [n_epochs x n_samples x '
                        'n_channels]')

    # Calculate freqs array
    freqs = np.linspace(0, fs/2, psd.shape[1])

    # Compute power
    psd_target_samp = \
        np.logical_and(freqs >= target_band[0], freqs <= target_band[1])
    psd_baseline_samp = \
        np.logical_and(freqs >= baseline_band[0], freqs <= baseline_band[1]) \
        if baseline_band is not None else np.ones_like(freqs).astype(int)
    band_power = np.sum(psd[:, psd_target_samp, :], axis=1) * \
                 (fs / (2 * freqs.shape[0]))
    baseline_power = np.sum(psd[:, psd_baseline_samp, :], axis=1) * \
                 (fs / (2 * freqs.shape[0]))

    return band_power / baseline_power


def median_frequency(psd, fs, target_band=(1, 70)):
    """This method computes the median frequency (MF) of the signal in the given
    band.

    Parameters
    ----------
    psd : numpy array
        PSD of the signal with shape [n_epochs, n_samples, n_channels]. Some
        of these dimensions may not exist in advance. In these case, create new
        axis using np.newaxis. E.g., non-epoched single-channel psd with
        shape [n_samples] can be passed to this function with psd[numpy.newaxis,
        ..., numpy.newaxis]. Afterwards, you may use numpy.squeeze to eliminate
        those axes.
    fs : int
        Sampling frequency of the signal
    target_band : numpy array
        Frequency band where the MF will be computed. [b1_start, b1_end].
        Default [1, 70].

    Returns
    -------
    median_freqs : numpy 2D array
        MF value for each epoch and channel with shape [n_epochs x n_channels].
    """
    # To numpy arrays
    psd = np.array(psd)
    target_band = np.array(target_band)

    # Check errors
    if len(psd.shape) != 3:
        raise ValueError('Parameter psd does not have correct dimensions. '
                         'Check the documentation for more information.')
    if len(target_band.shape) != 1 and target_band.shape[0] != 2:
        raise Exception('Parameter band must be an array with the desired '
                        'band. E.g., Delta: [0, 4]')

    # Calculate freqs array
    freqs = np.linspace(0, fs / 2, psd.shape[1])

    # Compute median frequency
    idx = np.logical_and(freqs >= target_band[0], freqs < target_band[1])
    freqs_in_band = freqs[idx]
    # Calculate total power
    total_power = np.sum(psd[:, idx, :], axis=1, keepdims=True)
    # Calculate cumulative power
    cum_power = np.cumsum(psd[:, idx, :], axis=1)
    # Get median frequency
    median_freq_idx = np.argmax(
        np.cumsum(cum_power <= (total_power / 2), axis=1), axis=1)
    median_freqs = freqs_in_band[median_freq_idx]

    return median_freqs


def shannon_spectral_entropy(psd, fs, target_band=(1, 70)):
    """Computes the Shannon spectral entropy of the power spectral density (PSD)
    in the given band.

   Parameters
    ----------
    psd : numpy array
        PSD of the signal with shape [n_epochs, n_samples, n_channels]. Some
        of these dimensions may not exist in advance. In these case, create new
        axis using np.newaxis. E.g., non-epoched single-channel psd with
        shape [n_samples] can be passed to this function with psd[numpy.newaxis,
        ..., numpy.newaxis]. Afterwards, you may use numpy.squeeze to eliminate
        those axes.
    fs : int
        Sampling frequency of the signal
    target_band : numpy array
        Frequency band where the SE will be computed. [b1_start, b1_end].
        Default [1, 70]

    Returns
    -------
    sample_entropy : numpy 2D array
        SE value for each epoch and channel with shape [n_epochs x n_channels].
    """
    # To numpy arrays
    psd = np.array(psd)
    target_band = np.array(target_band)

    # Check errors
    if len(psd.shape) != 3:
        raise ValueError('Parameter psd does not have correct dimensions. '
                         'Check the documentation for more information.')
    if len(target_band.shape) != 1 and target_band.shape[0] != 2:
        raise Exception('Parameter band must be an array with the desired '
                        'band. E.g., Delta: [0, 4]')

    # Calculate freqs array
    freqs = np.linspace(0, fs/2, psd.shape[1])

    # Compute shannon entropy
    idx = np.logical_and(freqs >= target_band[0], freqs < target_band[1])
    # Calculate total power
    total_power = np.sum(psd[:, idx, :], axis=1, keepdims=True)
    # Calculate the probability density function
    pdf = np.abs(psd[:, idx, :]) / total_power
    # Calculate shannon entropy
    se = -np.sum(pdf * np.log(pdf), axis=1) / np.log(pdf.shape[1])
    return se


# Built-in imports
import warnings
import os
# External imports
import tensorflow as tf
from tensorflow.keras.layers import Activation, Input, Flatten
from tensorflow.keras.layers import Dropout, BatchNormalization
from tensorflow.keras.layers import Conv2D, AveragePooling2D, DepthwiseConv2D
from tensorflow.keras.layers import Dense, SpatialDropout2D, SeparableConv2D
from tensorflow.keras.layers import Conv3D, AveragePooling3D, Add
from tensorflow.keras.constraints import max_norm
from tensorflow.keras.callbacks import EarlyStopping
import tensorflow.keras as keras
import sklearn.utils as sk_utils
import numpy as np

# Medusa imports
from medusa import components
from medusa import classification_utils
from medusa import tensorflow_integration


class EEGInceptionv1(components.ProcessingMethod):
    """EEG-Inception as described in Santamaría-Vázquez et al. 2020 [1]. This
    model is specifically designed for EEG classification tasks.

    References
    ----------
    [1] Santamaría-Vázquez, E., Martínez-Cagigal, V., Vaquerizo-Villar, F., &
    Hornero, R. (2020). EEG-Inception: A Novel Deep Convolutional Neural Network
    for Assistive ERP-based Brain-Computer Interfaces. IEEE Transactions on
    Neural Systems and Rehabilitation Engineering.
    """
    def __init__(self, input_time=1000, fs=128, n_cha=8, filters_per_branch=8,
                 scales_time=(500, 250, 125), dropout_rate=0.25,
                 activation='elu', n_classes=2, learning_rate=0.001,
                 gpu_acceleration=None):
        # Super call
        super().__init__(fit=[], predict_proba=['y_pred'])

        # Tensorflow config
        if gpu_acceleration is None:
            tensorflow_integration.check_tf_config(autoconfig=True)
        else:
            tensorflow_integration.config_tensorflow(gpu_acceleration)
        if tensorflow_integration.check_gpu_acceleration():
            os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
        tf.keras.backend.set_image_data_format('channels_last')

        # Parameters
        self.input_time = input_time
        self.fs = fs
        self.n_cha = n_cha
        self.filters_per_branch = filters_per_branch
        self.scales_time = scales_time
        self.dropout_rate = dropout_rate
        self.activation = activation
        self.n_classes = n_classes
        self.learning_rate = learning_rate

        # Useful variables
        self.input_samples = int(input_time * fs / 1000)
        self.scales_samples = [int(s * fs / 1000) for s in scales_time]

        # Create model
        self.model = self.__keras_model()
        # Create training callbacks
        self.training_callbacks = list()
        self.training_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.001,
                          mode='min',
                          patience=20,
                          verbose=1,
                          restore_best_weights=True)
        )
        # Create fine-tuning callbacks
        self.fine_tuning_callbacks = list()
        self.fine_tuning_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.0001,
                          mode='min',
                          patience=10,
                          verbose=1,
                          restore_best_weights=True)
        )

    def __keras_model(self):
        # ============================= INPUT ================================ #
        input_layer = Input((self.input_samples, self.n_cha, 1))
        # ================ BLOCK 1: SINGLE-CHANNEL ANALYSIS ================== #
        b1_units = list()
        for i in range(len(self.scales_samples)):
            unit = Conv2D(filters=self.filters_per_branch,
                          kernel_size=(self.scales_samples[i], 1),
                          kernel_initializer='he_normal',
                          padding='same')(input_layer)
            unit = BatchNormalization()(unit)
            unit = Activation(self.activation)(unit)
            unit = Dropout(self.dropout_rate)(unit)

            b1_units.append(unit)

        # Concatenation
        b1_out = keras.layers.concatenate(b1_units, axis=3)
        b1_out = AveragePooling2D((2, 1))(b1_out)

        # ================= BLOCK 2: SPATIAL FILTERING ======================= #
        b2_unit = DepthwiseConv2D((1, self.n_cha),
                                  use_bias=False,
                                  depth_multiplier=2,
                                  depthwise_constraint=max_norm(1.))(b1_out)
        b2_unit = BatchNormalization()(b2_unit)
        b2_unit = Activation(self.activation)(b2_unit)
        b2_unit = Dropout(self.dropout_rate)(b2_unit)
        b2_out = AveragePooling2D((2, 1))(b2_unit)

        # ================ BLOCK 3: MULTI-CHANNEL ANALYSIS =================== #
        b3_units = list()
        for i in range(len(self.scales_samples)):
            unit = Conv2D(filters=self.filters_per_branch,
                          kernel_size=(int(self.scales_samples[i] / 4), 1),
                          kernel_initializer='he_normal',
                          use_bias=False,
                          padding='same')(b2_out)
            unit = BatchNormalization()(unit)
            unit = Activation(self.activation)(unit)
            unit = Dropout(self.dropout_rate)(unit)

            b3_units.append(unit)

        # Concatenate + Average pooling
        b3_out = keras.layers.concatenate(b3_units, axis=3)
        b3_out = AveragePooling2D((2, 1))(b3_out)

        # ==================== BLOCK 4: OUTPUT-BLOCK ========================= #
        b4_u1 = Conv2D(filters=int(self.filters_per_branch *
                                   len(self.scales_samples) / 2),
                       kernel_size=(8, 1),
                       kernel_initializer='he_normal',
                       use_bias=False,
                       padding='same')(b3_out)
        b4_u1 = BatchNormalization()(b4_u1)
        b4_u1 = Activation(self.activation)(b4_u1)
        b4_u1 = AveragePooling2D((2, 1))(b4_u1)
        b4_u1 = Dropout(self.dropout_rate)(b4_u1)

        b4_u2 = Conv2D(filters=int(self.filters_per_branch *
                                   len(self.scales_samples) / 4),
                       kernel_size=(4, 1),
                       kernel_initializer='he_normal',
                       use_bias=False,
                       padding='same')(b4_u1)
        b4_u2 = BatchNormalization()(b4_u2)
        b4_u2 = Activation(self.activation)(b4_u2)
        b4_u2 = AveragePooling2D((2, 1))(b4_u2)
        b4_out = Dropout(self.dropout_rate)(b4_u2)

        # =========================== OUTPUT ================================= #
        # Output layer
        output_layer = Flatten()(b4_out)
        output_layer = Dense(self.n_classes, activation='softmax')(output_layer)
        # ============================ MODEL ================================= #
        # Optimizer
        optimizer = keras.optimizers.Adam(learning_rate=self.learning_rate,
                                          beta_1=0.9, beta_2=0.999,
                                          amsgrad=False)
        # Create and compile model
        model = keras.models.Model(inputs=input_layer,
                                   outputs=output_layer)
        model.compile(loss='categorical_crossentropy',
                      optimizer=optimizer,
                      metrics=['accuracy'])

        return model

    @staticmethod
    def transform_data(X, y=None):
        """Transforms input data to the correct dimensions for EEG-Inception

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Labels array. If labels are in categorical format, they will be
            converted to one-hot encoding.
        """
        if len(X.shape) == 3 or X.shape[-1] != 1:
            X = X.reshape((X.shape[0], X.shape[1], X.shape[2], 1))
        if y is None:
            return X
        else:
            if len(y.shape) == 1 or y.shape[-1] == 1:
                y = classification_utils.one_hot_labels(y)
            return X, y

    def fit(self, X, y, fine_tuning=False, shuffle_before_fit=False, **kwargs):
        """Fit the model. All additional keras parameters of class
        tensorflow.keras.Model will pass through. See keras documentation to
        know what can you do: https://keras.io/api/models/model_training_apis/.

        If no parameters are specified, some default options are set [1]:

            - Epochs: 100 if fine_tuning else 500
            - Batch size: 32 if fine_tuning else 1024
            - Callbacks: self.fine_tuning_callbacks if fine_tuning else
                self.training_callbacks

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        fine_tuning: bool
            Set to True to use the default training parameters for fine
            tuning. False by default.
        shuffle_before_fit: bool
            If True, the data will be shuffled before training just once. Note
            that if you use the keras native argument 'shuffle', the data is
            shuffled each epoch.
        kwargs:
            Key-value arguments will be passed to the fit function of the model.
            This way, you can set your own training parameters using keras API.
            See https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
        """
        # Check numpy arrays
        X = np.array(X)
        y = np.array(y)
        # Check GPU
        if not tensorflow_integration.check_gpu_acceleration():
            warnings.warn('GPU acceleration is not available. The training '
                          'time is drastically reduced with GPU.')
        # Shuffle the data before fitting
        if shuffle_before_fit:
            X, y = sk_utils.shuffle(X, y)
        # Training parameters
        if not fine_tuning:
            # Rewrite default values
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 500
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 1024
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.training_callbacks
        else:
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 100
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 32
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.fine_tuning_callbacks

        # Transform data if necessary
        X, y = self.transform_data(X, y)
        # Fit
        with tf.device(tensorflow_integration.get_tf_device_name()):
            return self.model.fit(X, y, **kwargs)

    def predict_proba(self, X):
        """Model prediction scores for the given features.

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        """
        # Check numpy arrays
        X = np.array(X)
        # Transform data if necessary
        X = self.transform_data(X)
        # Predict
        with tf.device(tensorflow_integration.get_tf_device_name()):
            return self.model.predict(X)

    def to_pickleable_obj(self):
        # Parameters
        kwargs = {
            'input_time': self.input_time,
            'fs': self.fs,
            'n_cha': self.n_cha,
            'filters_per_branch': self.filters_per_branch,
            'scales_time': self.scales_time,
            'dropout_rate': self.dropout_rate,
            'activation': self.activation,
            'n_classes': self.n_classes,
            'learning_rate': self.learning_rate,
        }
        weights = self.model.get_weights()
        # Pickleable object
        pickleable_obj = {
            'kwargs': kwargs,
            'weights': weights
        }
        return pickleable_obj

    @classmethod
    def from_pickleable_obj(cls, pickleable_obj):
        model = cls(**pickleable_obj['kwargs'])
        model.model.set_weights(pickleable_obj['weights'])
        return model

    def set_weights(self, weights):
        return self.model.set_weights(weights)

    def get_weights(self):
        return self.model.get_weights()

    def save_weights(self, path):
        return self.model.save_weights(path)

    def load_weights(self, weights_path):
        return self.model.load_weights(weights_path)


class EEGNet(components.ProcessingMethod):
    """EEG-Inception as described in Lawhern et al. 2018 [1]. This model is
    specifically designed for EEG classification tasks.

    Original source https://github.com/vlawhern/arl-eegmodels

    References
    ----------
    [1] Lawhern, V. J., Solon, A. J., Waytowich, N. R., Gordon, S. M., Hung,
    C. P., & Lance, B. J. (2018). EEGNet: a compact convolutional neural network
    for EEG-based brain–computer interfaces. Journal of neural engineering,
    15(5), 056013.
    """
    def __init__(self, nb_classes, n_cha=64, samples=128, dropout_rate=0.5,
                 kern_length=64, F1=8, D=2, F2=16, norm_rate=0.25,
                 dropout_type='Dropout', learning_rate=0.001,
                 gpu_acceleration=None):

        # Super call
        super().__init__(fit=[], predict_proba=['y_pred'])

        # Tensorflow config
        if gpu_acceleration is None:
            tensorflow_integration.check_tf_config(autoconfig=True)
        else:
            tensorflow_integration.config_tensorflow(gpu_acceleration)
        if tensorflow_integration.check_gpu_acceleration():
            os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
        tf.keras.backend.set_image_data_format('channels_last')

        # Parameters
        self.nb_classes = nb_classes
        self.n_cha = n_cha
        self.samples = samples
        self.dropout_rate = dropout_rate
        self.kern_length = kern_length
        self.F1 = F1
        self.D = D
        self.F2 = F2
        self.norm_rate = norm_rate
        self.dropout_type = dropout_type
        self.learning_rate = learning_rate

        # Create model
        self.model = self.__keras_model()

        # Create training callbacks
        self.training_callbacks = list()
        self.training_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.001,
                          mode='min',
                          patience=20,
                          verbose=1,
                          restore_best_weights=True)
        )
        # Create fine-tuning callbacks
        self.fine_tuning_callbacks = list()
        self.fine_tuning_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.0001,
                          mode='min',
                          patience=10,
                          verbose=1,
                          restore_best_weights=True)
        )

    def __keras_model(self):
        """ Keras Implementation of EEGNet
        http://iopscience.iop.org/article/10.1088/1741-2552/aace8c/meta
        Note that this implements the newest version of EEGNet and NOT the
        earlier version (version v1 and v2 on arxiv). We strongly recommend
        using this architecture as it performs much better and has nicer
        properties than our earlier version. For example:

            1. Depthwise Convolutions to learn spatial filters within a
            temporal convolution. The use of the depth_multiplier option maps
            exactly to the number of spatial filters learned within a temporal
            filter. This matches the setup of algorithms like FBCSP which learn
            spatial filters within each filter in a filter-bank. This also
            limits the number of free parameters to fit when compared to a
            fully-connected convolution.

            2. Separable Convolutions to learn how to optimally combine spatial
            filters across temporal bands. Separable Convolutions are Depthwise
            Convolutions followed by (1x1) Pointwise Convolutions.


        While the original paper used Dropout, we found that SpatialDropout2D
        sometimes produced slightly better results for classification of ERP
        signals. However, SpatialDropout2D significantly reduced performance
        on the Oscillatory dataset (SMR, BCI-IV Dataset 2A). We recommend using
        the default Dropout in most cases.

        Assumes the input signal is sampled at 128Hz. If you want to use this
        model for any other sampling rate you will need to modify the lengths of
        temporal kernels and average pooling size in blocks 1 and 2 as needed
        (double the kernel lengths for double the sampling rate, etc). Note
        that we haven't tested the model performance with this rule so this
        may not work well.

        The model with default parameters gives the EEGNet-8,2 model as
        discussed in the paper. This model should do pretty well in general,
        although it is advised to do some model searching to get optimal
        performance on your particular dataset.

        We set F2 = F1 * D (number of input filters = number of output filters)
        for the SeparableConv2D layer. We haven't extensively tested other
        values of this parameter (say, F2 < F1 * D for compressed learning,
        and F2 > F1 * D for overcomplete). We believe the main parameters to
        focus on are F1 and D.

        Inputs:

          nb_classes      : int, number of classes to classify
          Chans, Samples  : number of channels and time points in the EEG data
          dropoutRate     : dropout fraction
          kernLength      : length of temporal convolution in first layer. We
                            found that setting this to be half the sampling
                            rate worked well in practice. For the SMR dataset in
                            particular since the data was high-passed at 4Hz
                            we used a kernel length of 32.
          F1, F2          : number of temporal filters (F1) and number of
                            pointwise filters (F2) to learn. Default: F1 = 8,
                            F2 = F1 * D.
          D               : number of spatial filters to learn within each
                            temporal convolution. Default: D = 2
          dropoutType     : Either SpatialDropout2D or Dropout, passed as a
                            string.
        """

        if self.dropout_type == 'SpatialDropout2D':
            dropout_type = SpatialDropout2D
        elif self.dropout_type == 'Dropout':
            dropout_type = Dropout
        else:
            raise ValueError('dropoutType must be one of SpatialDropout2D '
                             'or Dropout, passed as a string.')

        input1 = Input(shape=(self.n_cha, self.samples, 1))

        ##################################################################
        block1 = Conv2D(self.F1, (1, self.kern_length), padding='same',
                        input_shape=(self.n_cha, self.samples, 1),
                        use_bias=False)(input1)
        block1 = BatchNormalization()(block1)
        block1 = DepthwiseConv2D((self.n_cha, 1), use_bias=False,
                                 depth_multiplier=self.D,
                                 depthwise_constraint=max_norm(1.))(block1)
        block1 = BatchNormalization()(block1)
        block1 = Activation('elu')(block1)
        block1 = AveragePooling2D((1, 4))(block1)
        block1 = dropout_type(self.dropout_rate)(block1)

        block2 = SeparableConv2D(self.F2, (1, 16),
                                 use_bias=False, padding='same')(block1)
        block2 = BatchNormalization()(block2)
        block2 = Activation('elu')(block2)
        block2 = AveragePooling2D((1, 8))(block2)
        block2 = dropout_type(self.dropout_rate)(block2)

        flatten = Flatten(name='flatten')(block2)

        dense = Dense(self.nb_classes, name='dense',
                      kernel_constraint=max_norm(self.norm_rate))(flatten)
        softmax = Activation('softmax', name='softmax')(dense)

        # Create and compile model
        model = keras.models.Model(inputs=input1, outputs=softmax)
        optimizer = keras.optimizers.Adam(learning_rate=self.learning_rate,
                                          beta_1=0.9, beta_2=0.999,
                                          amsgrad=False)
        model.compile(loss='categorical_crossentropy',
                      optimizer=optimizer,
                      metrics=['accuracy'])

        return model

    @staticmethod
    def transform_data(X, y=None):
        """Transforms input data to the correct dimensions for EEG-Inception

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Labels array. If labels are in categorical format, they will be
            converted to one-hot encoding.
        """
        if len(X.shape) == 3 or X.shape[-1] != 1:
            X = X.reshape((X.shape[0], X.shape[1], X.shape[2], 1))
            X = np.swapaxes(X, 1, 2)
        if y is None:
            return X
        else:
            if len(y.shape) == 1 or y.shape[-1] == 1:
                y = classification_utils.one_hot_labels(y)
            return X, y

    def fit(self, X, y, fine_tuning=False, shuffle_before_fit=False, **kwargs):
        """Fit the model. All additional keras parameters of class
        tensorflow.keras.Model will pass through. See keras documentation to
        know what can you do: https://keras.io/api/models/model_training_apis/.

        If no parameters are specified, some default options are set [1]:

            - Epochs: 100 if fine_tuning else 500
            - Batch size: 32 if fine_tuning else 1024
            - Callbacks: self.fine_tuning_callbacks if fine_tuning else
                self.training_callbacks

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        fine_tuning: bool
            Set to True to use the default training parameters for fine
            tuning. False by default.
        shuffle_before_fit: bool
            If True, the data will be shuffled before training.
        kwargs:
            Key-value arguments will be passed to the fit function of the model.
            This way, you can set your own training parameters for keras.
            See https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
        """
        # Check numpy arrays
        X = np.array(X)
        y = np.array(y)
        # Check GPU
        if not tensorflow_integration.check_gpu_acceleration():
            warnings.warn('GPU acceleration is not available. The training '
                          'time is drastically reduced with GPU.')

        # Shuffle the data before fitting
        if shuffle_before_fit:
            X, y = sk_utils.shuffle(X, y)

        # Training parameters
        if not fine_tuning:
            # Rewrite default values
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 500
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 1024
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.training_callbacks
        else:
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 100
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 32
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.fine_tuning_callbacks

        # Transform data if necessary
        X, y = self.transform_data(X, y)
        # Fit
        with tf.device(tensorflow_integration.get_tf_device_name()):
            return self.model.fit(X, y, **kwargs)

    def predict_proba(self, X):
        """Model prediction scores for the given features.

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        """
        # Check numpy arrays
        X = np.array(X)
        # Transform data if necessary
        X = self.transform_data(X)
        # Predict
        with tf.device(tensorflow_integration.get_tf_device_name()):
            return self.model.predict(X)

    def to_pickleable_obj(self):
        # Key data
        kwargs = {
            'nb_classes': self.nb_classes,
            'n_cha': self.n_cha,
            'samples': self.samples,
            'dropout_rate': self.dropout_rate,
            'kern_length': self.kern_length,
            'F1': self.F1,
            'D': self.D,
            'F2': self.F2,
            'norm_rate': self.norm_rate,
            'dropout_type': self.dropout_type,
            'learning_rate': self.learning_rate
        }
        weights = self.model.get_weights()
        # Pickleable object
        pickleable_obj = {
            'kwargs': kwargs,
            'weights': weights
        }
        return pickleable_obj

    @classmethod
    def from_pickleable_obj(cls, pickleable_obj):
        model = cls(**pickleable_obj['kwargs'])
        model.model.set_weights(pickleable_obj['weights'])
        return model

    def set_weights(self, weights):
        return self.model.set_weights(weights)

    def get_weights(self):
        return self.model.get_weights()

    def save_weights(self, path):
        return self.model.save_weights(path)

    def load_weights(self, weights_path):
        return self.model.load_weights(weights_path)


class EEGSym(components.ProcessingMethod):
    """EEGSym as described in Pérez-Velasco et al. 2022 [1]. This
    model is specifically designed for EEG classification tasks.

    References
    ----------
    [1] Pérez-Velasco, S., Santamaría-Vázquez, E., Martínez-Cagigal, V.,
    Marcos-Martínez, D., & Hornero, R. (2022). EEGSym: Overcoming Intersubject
    Variability in Motor Imagery Based BCIs with Deep Learning. IEEE
    Transactions on Neural Systems and Rehabilitation Engineering.
    """
    #TODO: Implement automatic ordering of channels
    #TODO: Implement trial iterator and data augmentation
    def __init__(self, input_time=3000, fs=128, n_cha=8, filters_per_branch=24,
           scales_time=(500, 250, 125), dropout_rate=0.4, activation='elu',
           n_classes=2, learning_rate=0.001, ch_lateral=3,
           spatial_resnet_repetitions=1, residual=True, symmetric=True,
                 gpu_acceleration=None):
        # Super call
        super().__init__(fit=[], predict_proba=['y_pred'])

        # Tensorflow config
        if gpu_acceleration is None:
            tensorflow_integration.check_tf_config(autoconfig=True)
        else:
            tensorflow_integration.config_tensorflow(gpu_acceleration)
        if tensorflow_integration.check_gpu_acceleration():
            os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
        tf.keras.backend.set_image_data_format('channels_last')

        # Parameters
        self.input_time = input_time
        self.fs = fs
        self.n_cha = n_cha
        self.filters_per_branch = filters_per_branch
        self.scales_time = scales_time
        self.dropout_rate = dropout_rate
        self.activation = activation
        self.n_classes = n_classes
        self.learning_rate = learning_rate
        self.ch_lateral = ch_lateral
        self.spatial_resnet_repetitions = spatial_resnet_repetitions
        self.residual = residual
        self.symmetric = symmetric
        self.initialized = False

        # Useful variables
        self.input_samples = int(input_time * fs / 1000)
        self.scales_samples = [int(s * fs / 1000) for s in scales_time]

        # Create model
        self.model = self.__keras_model()
        # Create training callbacks
        self.training_callbacks = list()
        self.training_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.001,
                          mode='min',
                          patience=20,
                          verbose=1,
                          restore_best_weights=True)
        )
        # Create fine-tuning callbacks
        self.fine_tuning_callbacks = list()
        self.fine_tuning_callbacks.append(
            EarlyStopping(monitor='val_loss',
                          min_delta=0.0001,
                          mode='min',
                          patience=10,
                          verbose=1,
                          restore_best_weights=True)
        )

    def __keras_model(self):
        # ============================= INPUT ================================ #
        # ==================================================================== #
        input_layer = Input((self.input_samples, self.n_cha, 1))
        input = tf.expand_dims(input_layer, axis=1)
        ncha = self.n_cha
        if self.symmetric:
            superposition = False
            if self.ch_lateral < self.n_cha // 2:
                superposition = True
            ncha = self.n_cha - self.ch_lateral

            left_idx = list(range(self.ch_lateral))
            ch_left = tf.gather(input, indices=left_idx, axis=-2)
            right_idx = list(np.array(left_idx) + int(ncha))
            ch_right = tf.gather(input, indices=right_idx, axis=-2)

            if superposition:
                central_idx = list(
                    np.array(range(ncha - self.ch_lateral)) +
                    self.ch_lateral)
                ch_central = tf.gather(input, indices=central_idx, axis=-2)

                left_init = keras.layers.concatenate((ch_left, ch_central),
                                                     axis=-2)
                right_init = keras.layers.concatenate((ch_right, ch_central),
                                                      axis=-2)
            else:
                left_init = ch_left
                right_init = ch_right

            input = keras.layers.concatenate((left_init, right_init), axis=1)
            division = 2
        else:
            division = 1
        # ================ GENERAL INCEPTION/RESIDUAL MODULE ================= #
        # ==================================================================== #

        # ====================== TEMPOSPATIAL ANALYSIS ======================= #
        # ==================================================================== #
        # ========================== Inception (x2) ========================== #
        b1_out = self.general_module([input],
                                scales_samples=self.scales_samples,
                                filters_per_branch=self.filters_per_branch,
                                ncha=ncha,
                                activation=self.activation,
                                dropout_rate=self.dropout_rate, average=2,
                                spatial_resnet_repetitions=
                                self.spatial_resnet_repetitions,
                                residual=self.residual, init=True)

        b2_out = self.general_module(b1_out, scales_samples=[int(x / 4) for x in
                                                        self.scales_samples],
                                filters_per_branch=self.filters_per_branch,
                                ncha=ncha,
                                activation=self.activation,
                                dropout_rate=self.dropout_rate, average=2,
                                spatial_resnet_repetitions=
                                self.spatial_resnet_repetitions,
                                residual=self.residual)
        # ============================== Residual (x3) =========================== #
        b3_u1 = self.general_module(b2_out, scales_samples=[16],
                               filters_per_branch=int(
                                   self.filters_per_branch * len(
                                       self.scales_samples) / 2),
                               ncha=ncha,
                               activation=self.activation,
                               dropout_rate=self.dropout_rate, average=2,
                               spatial_resnet_repetitions=
                               self.spatial_resnet_repetitions,
                               residual=self.residual)
        b3_u1 = self.general_module(b3_u1,
                               scales_samples=[8],
                               filters_per_branch=int(
                                   self.filters_per_branch * len(
                                       self.scales_samples) / 2),
                               ncha=ncha,
                               activation=self.activation,
                               dropout_rate=self.dropout_rate, average=2,
                               spatial_resnet_repetitions=
                               self.spatial_resnet_repetitions,
                               residual=self.residual)
        b3_u2 = self.general_module(b3_u1, scales_samples=[4],
                               filters_per_branch=int(
                                   self.filters_per_branch * len(
                                       self.scales_samples) / 4),
                               ncha=ncha,
                               activation=self.activation,
                               dropout_rate=self.dropout_rate, average=2,
                               spatial_resnet_repetitions=
                               self.spatial_resnet_repetitions,
                               residual=self.residual)

        t_red = b3_u2[0]
        for _ in range(1):
            t_red_temp = t_red
            t_red_temp = Conv3D(kernel_size=(1, 4, 1),
                                filters=int(self.filters_per_branch * len(
                                    self.scales_samples) / 4),
                                use_bias=False,
                                strides=(1, 1, 1),
                                kernel_initializer='he_normal',
                                padding='same')(t_red_temp)
            t_red_temp = BatchNormalization()(t_red_temp)
            t_red_temp = Activation(self.activation)(t_red_temp)
            t_red_temp = Dropout(self.dropout_rate)(t_red_temp)

            if self.residual:
                t_red = Add()([t_red, t_red_temp])
            else:
                t_red = t_red_temp

        t_red = AveragePooling3D((1, 2, 1))(t_red)
        # ========================= CHANNEL MERGING ========================== #
        # ==================================================================== #
        ch_merg = t_red
        if self.residual:
            for _ in range(2):
                ch_merg_temp = ch_merg
                ch_merg_temp = Conv3D(kernel_size=(division, 1, ncha),
                                      filters=int(self.filters_per_branch * len(
                                          self.scales_samples) / 4),
                                      use_bias=False,
                                      strides=(1, 1, 1),
                                      kernel_initializer='he_normal',
                                      padding='valid')(ch_merg_temp)
                ch_merg_temp = BatchNormalization()(ch_merg_temp)
                ch_merg_temp = Activation(self.activation)(ch_merg_temp)
                ch_merg_temp = Dropout(self.dropout_rate)(ch_merg_temp)

                ch_merg = Add()([ch_merg, ch_merg_temp])

            ch_merg = Conv3D(kernel_size=(division, 1, ncha),
                             filters=int(
                                 self.filters_per_branch * len(
                                     self.scales_samples) / 4),
                             groups=int(
                                 self.filters_per_branch * len(
                                     self.scales_samples) / 8),
                             use_bias=False,
                             padding='valid')(ch_merg)
            ch_merg = BatchNormalization()(ch_merg)
            ch_merg = Activation(self.activation)(ch_merg)
            ch_merg = Dropout(self.dropout_rate)(ch_merg)
        else:
            if self.symmetric:
                ch_merg = Conv3D(kernel_size=(division, 1, 1),
                                 filters=int(
                                     self.filters_per_branch * len(
                                         self.scales_samples) / 4),
                                 groups=int(
                                     self.filters_per_branch * len(
                                         self.scales_samples) / 8),
                                 use_bias=False,
                                 padding='valid')(ch_merg)
                ch_merg = BatchNormalization()(ch_merg)
                ch_merg = Activation(self.activation)(ch_merg)
                ch_merg = Dropout(self.dropout_rate)(ch_merg)
        # ========================== TEMPORAL MERGING ============================ #
        # ======================================================================== #
        t_merg = ch_merg
        for _ in range(1):
            if self.residual:
                t_merg_temp = t_merg
                t_merg_temp = Conv3D(kernel_size=(1, self.input_samples // 64,
                                                  1),
                                     filters=int(self.filters_per_branch * len(
                                         self.scales_samples) / 4),
                                     use_bias=False,
                                     strides=(1, 1, 1),
                                     kernel_initializer='he_normal',
                                     padding='valid')(t_merg_temp)
                t_merg_temp = BatchNormalization()(t_merg_temp)
                t_merg_temp = Activation(self.activation)(t_merg_temp)
                t_merg_temp = Dropout(self.dropout_rate)(t_merg_temp)

                t_merg = Add()([t_merg, t_merg_temp])
            else:
                t_merg_temp = t_merg
                t_merg_temp = Conv3D(kernel_size=(1, self.input_samples // 64, 1),
                                     filters=int(self.filters_per_branch * len(
                                         self.scales_samples) / 4),
                                     use_bias=False,
                                     strides=(1, 1, 1),
                                     kernel_initializer='he_normal',
                                     padding='same')(t_merg_temp)
                t_merg_temp = BatchNormalization()(t_merg_temp)
                t_merg_temp = Activation(self.activation)(t_merg_temp)
                t_merg_temp = Dropout(self.dropout_rate)(t_merg_temp)
                t_merg = t_merg_temp

        t_merg = Conv3D(kernel_size=(1, self.input_samples // 64, 1),
                        filters=int(
                            self.filters_per_branch * len(
                                self.scales_samples) / 4) * 2,
                        groups=int(
                            self.filters_per_branch * len(
                                self.scales_samples) / 4),
                        use_bias=False,
                        padding='valid')(t_merg)
        t_merg = BatchNormalization()(t_merg)
        t_merg = Activation(self.activation)(t_merg)
        t_merg = Dropout(self.dropout_rate)(t_merg)
        # =============================== OUTPUT ================================= #
        output = t_merg
        for _ in range(4):
            output_temp = output
            output_temp = Conv3D(kernel_size=(1, 1, 1),
                                 filters=int(self.filters_per_branch * len(
                                     self.scales_samples) / 2),
                                 use_bias=False,
                                 strides=(1, 1, 1),
                                 kernel_initializer='he_normal',
                                 padding='valid')(output_temp)
            output_temp = BatchNormalization()(output_temp)
            output_temp = Activation(self.activation)(output_temp)
            output_temp = Dropout(self.dropout_rate)(output_temp)
            if self.residual:
                output = Add()([output, output_temp])
            else:
                output = output_temp
        output = Flatten()(output)
        output_layer = Dense(self.n_classes, activation='softmax')(output)
        # ============================ MODEL ================================= #
        # Optimizer
        optimizer = keras.optimizers.Adam(learning_rate=self.learning_rate,
                                          beta_1=0.9, beta_2=0.999,
                                          amsgrad=False)
        # Create and compile model
        model = keras.models.Model(inputs=input_layer,
                                   outputs=output_layer)
        model.compile(loss='categorical_crossentropy',
                      optimizer=optimizer,
                      metrics=['accuracy'])

        return model

    @staticmethod
    def general_module(input, scales_samples, filters_per_branch, ncha,
                       activation, dropout_rate, average,
                       spatial_resnet_repetitions=1, residual=True,
                       init=False):
        """General inception/residual module.

            This function returns the input with the operations of a
            inception or residual module from the publication applied.

            Parameters
            ----------
            input : list
                List of input blocks to the module.
            scales_samples : list
                List of samples size of the temporal operations kernels.
            filters_per_branch : int
                Number of filters in each Inception branch. The number should be
                multiplies of 8.
            ncha :
                Number of input channels.
            activation : str
                Activation
            dropout_rate : float
                Dropout rate
            spatial_resnet_repetitions: int
                Number of repetitions of the operations of spatial analysis at
                each step of the spatiotemporal analysis. In the original
                publication this value was set to 1 and not tested its
                variations.
            residual : Bool
                If the residual operations are present in EEGSym architecture.
            init : Bool
                If the module is the first one applied to the input, to apply a
                channel merging operation if the architecture does not include
                residual operations.

            Returns
            -------
            block_out : list
                List of outputs modules
        """
        block_units = list()
        unit_conv_t = list()
        unit_batchconv_t = list()

        for i in range(len(scales_samples)):
            unit_conv_t.append(Conv3D(filters=filters_per_branch,
                                      kernel_size=(1, scales_samples[i], 1),
                                      kernel_initializer='he_normal',
                                      padding='same'))
            unit_batchconv_t.append(BatchNormalization())

        if ncha != 1:
            unit_dconv = list()
            unit_batchdconv = list()
            unit_conv_s = list()
            unit_batchconv_s = list()
            for i in range(spatial_resnet_repetitions):
                # 3D Implementation of DepthwiseConv
                unit_dconv.append(Conv3D(kernel_size=(1, 1, ncha),
                                         filters=filters_per_branch * len(
                                             scales_samples),
                                         groups=filters_per_branch * len(
                                             scales_samples),
                                         use_bias=False,
                                         padding='valid'))
                unit_batchdconv.append(BatchNormalization())

                unit_conv_s.append(Conv3D(kernel_size=(1, 1, ncha),
                                          filters=filters_per_branch,
                                          # groups=filters_per_branch,
                                          use_bias=False,
                                          strides=(1, 1, 1),
                                          kernel_initializer='he_normal',
                                          padding='valid'))
                unit_batchconv_s.append(BatchNormalization())

            unit_conv_1 = Conv3D(kernel_size=(1, 1, 1),
                                 filters=filters_per_branch,
                                 use_bias=False,
                                 kernel_initializer='he_normal',
                                 padding='valid')
            unit_batchconv_1 = BatchNormalization()

        # Temporal analysis stage of the module
        for j in range(len(input)):
            block_side_units = list()
            for i in range(len(scales_samples)):
                unit = input[j]
                unit = unit_conv_t[i](unit)

                unit = unit_batchconv_t[i](unit)
                unit = Activation(activation)(unit)
                unit = Dropout(dropout_rate)(unit)

                block_side_units.append(unit)
            block_units.append(block_side_units)
        # Concatenation of possible inception modules
        block_out = list()
        for j in range(len(input)):
            if len(block_units[j]) != 1:
                block_out.append(
                    keras.layers.concatenate(block_units[j], axis=-1))
                block_out_temp = input[j]
            else:
                block_out.append(block_units[j][0])
                block_out_temp = input[j]
                block_out_temp = unit_conv_1(block_out_temp)
                block_out_temp = unit_batchconv_1(block_out_temp)
                block_out_temp = Activation(activation)(block_out_temp)
                block_out_temp = Dropout(dropout_rate)(block_out_temp)

            if residual:
                block_out[j] = Add()([block_out[j], block_out_temp])
            else:
                block_out[j] = block_out_temp

            if average != 1:
                block_out[j] = AveragePooling3D((1, average, 1))(block_out[j])

        # Spatial analysis stage of the module
        if ncha != 1:
            for i in range(spatial_resnet_repetitions):
                block_out_temp = list()
                for j in range(len(input)):
                    if len(scales_samples) != 1:
                        if residual:
                            block_out_temp.append(block_out[j])

                            block_out_temp[j] = unit_dconv[i](block_out_temp[j])

                            block_out_temp[j] = unit_batchdconv[i](
                                block_out_temp[j])
                            block_out_temp[j] = Activation(activation)(
                                block_out_temp[j])
                            block_out_temp[j] = Dropout(dropout_rate)(
                                block_out_temp[j])

                            block_out[j] = Add()(
                                [block_out[j], block_out_temp[j]])

                        elif init:
                            block_out[j] = unit_dconv[i](block_out[j])
                            block_out[j] = unit_batchdconv[i](block_out[j])
                            block_out[j] = Activation(activation)(block_out[j])
                            block_out[j] = Dropout(dropout_rate)(block_out[j])
                    else:
                        if residual:
                            block_out_temp.append(block_out[j])

                            block_out_temp[j] = unit_conv_s[i](
                                block_out_temp[j])
                            block_out_temp[j] = unit_batchconv_s[i](
                                block_out_temp[j])
                            block_out_temp[j] = Activation(activation)(
                                block_out_temp[j])
                            block_out_temp[j] = Dropout(dropout_rate)(
                                block_out_temp[j])

                            block_out[j] = Add()(
                                [block_out[j], block_out_temp[j]])
        return block_out

    @staticmethod
    def transform_data(X, y=None):
        """Transforms input data to the correct dimensions for EEGSym

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Labels array. If labels are in categorical format, they will be
            converted to one-hot encoding.
        """
        if len(X.shape) == 3 or X.shape[-1] != 1:
            X = X.reshape((X.shape[0], X.shape[1], X.shape[2], 1))
        if y is None:
            return X
        else:
            if len(y.shape) == 1 or y.shape[-1] == 1:
                y = classification_utils.one_hot_labels(y)
            return X, y

    def preprocessing_function(self, augmentation=True):
        """Custom Data Augmentation for EEGSym.

            Parameters
            ----------
            augmentation : Bool
                If the augmentation is performed to the input.

            Returns
            -------
            data_augmentation : function
                Data augmentation performed to each trial
        """

        def data_augmentation(trial):
            """Custom Data Augmentation for EEGSym.

                Parameters
                ----------
                trial : tf.tensor
                    Input of the

                Returns
                -------
                data_augmentation : keras.models.Model
                    Data augmentation performed to each trial
            """

            samples, ncha, _ = trial.shape

            augmentations = dict()
            augmentations["patch_perturbation"] = 0
            augmentations["random_shift"] = 0
            augmentations["hemisphere_perturbation"] = 0
            augmentations["no_augmentation"] = 0

            # selectionables = ["patch_perturbation", "random_shift",
            #                   "hemisphere_perturbation", "no_augmentation"]
            # We eliminate hemisphere_perturbation due to it being very
            # dependant on the channels introduced
            selectionables = ["patch_perturbation", "random_shift",
                              "no_augmentation"]
            probabilities = None

            if augmentation:
                selection = np.random.choice(selectionables, p=probabilities)
                augmentations[selection] = 1

                method = np.random.choice((0, 2))
                std = 'self'
                # elif data_augmentation == 1:  # Random shift
                for _ in range(augmentations["random_shift"]):  # Random shift
                    # Select position where to erase that timeframe
                    position = 0
                    if position == 0:
                        samples_shifted = np.random.randint(low=1, high=int(
                            samples * 0.5 / 3))
                    else:
                        samples_shifted = np.random.randint(low=1, high=int(
                            samples * 0.1 / 3))

                    if method == 0:
                        shifted_samples = np.zeros((samples_shifted, ncha, 1))
                    else:
                        if std == 'self':
                            std_applied = np.std(trial)
                        else:
                            std_applied = std
                        center = 0
                        shifted_samples = np.random.normal(center, std_applied,
                                                           (samples_shifted,
                                                            ncha,
                                                            1))
                    if position == 0:
                        trial = np.concatenate((shifted_samples, trial),
                                               axis=0)[:samples]
                    else:
                        trial = np.concatenate((trial, shifted_samples),
                                               axis=0)[samples_shifted:]

                for _ in range(
                        augmentations[
                            "patch_perturbation"]):  # Patch perturbation
                    channels_affected = np.random.randint(low=1, high=ncha - 1)
                    pct_max = 1
                    pct_min = 0.2
                    pct_erased = np.random.uniform(low=pct_min, high=pct_max)
                    # Select time to be erased acording to pct_erased
                    # samples_erased = np.min((int(samples*ncha*pct_erased//channels_affected), samples))#np.random.randint(low=1, high=samples//3)
                    samples_erased = int(samples * pct_erased)
                    # Select position where to erase that timeframe
                    if samples_erased != samples:
                        samples_idx = np.arange(
                            samples_erased) + np.random.randint(
                            samples - samples_erased)
                    else:
                        samples_idx = np.arange(samples_erased)
                    # Select indexes to erase (always keep at least a channel)
                    channel_idx = np.random.permutation(np.arange(ncha))[
                                  :channels_affected]
                    channel_idx.sort()
                    for channel in channel_idx:
                        if method == 0:
                            trial[samples_idx, channel] = 0
                        else:
                            if std == 'self':
                                std_applied = np.std(trial[:, channel]) \
                                              * np.random.uniform(low=0.01,
                                                                  high=2)
                            else:
                                std_applied = std
                            center = 0
                            trial[samples_idx, channel] += \
                                np.random.normal(center, std_applied,
                                                 trial[samples_idx, channel,
                                                 :].shape)
                            # Standarize the channel again after the change
                            temp_trial_ch_mean = np.mean(trial[:, channel],
                                                         axis=0)
                            temp_trial_ch_std = np.std(trial[:, channel],
                                                       axis=0)
                            trial[:, channel] = (trial[:,
                                                 channel] - temp_trial_ch_mean) / temp_trial_ch_std

                for _ in range(augmentations["hemisphere_perturbation"]):
                    # Select side to mix/change for noise
                    left_right = np.random.choice((0, 1))
                    if method == 0:
                        if left_right == 1:
                            channel_idx = np.arange(ncha)[:int((ncha / 2) - 1)]
                            channel_mix = np.random.permutation(
                                channel_idx.copy())
                        else:
                            channel_idx = np.arange(ncha)[-int((ncha / 2) - 1):]
                            channel_mix = np.random.permutation(
                                channel_idx.copy())
                        temp_trial = trial.copy()
                        for channel, channel_mixed in zip(channel_idx,
                                                          channel_mix):
                            temp_trial[:, channel] = trial[:, channel_mixed]
                        trial = temp_trial
                    else:
                        if left_right == 1:
                            channel_idx = np.arange(ncha)[:int((ncha / 2) - 1)]
                        else:
                            channel_idx = np.arange(ncha)[-int((ncha / 2) - 1):]
                        for channel in channel_idx:
                            trial[:, channel] = np.random.normal(0, 1,
                                                                 trial[:,
                                                                 channel].shape)

            return trial

        return data_augmentation

    def trial_iterator(self, X, y, batch_size=32, shuffle=True,
                       augmentation=True):
        """Custom trial iterator to pretrain EEGSym.

            Parameters
            ----------
            X : tf.tensor
                Input tensor of  EEG features.
            y : tf.tensor
                Input tensor of  labels.
            batch_size : int
                Number of features in each batch.
            shuffle : Bool
                If the features are shuffled at each training epoch.
            augmentation : Bool
                If the augmentation is performed to the input.

            Returns
            -------
            trial_iterator : tf.keras.preprocessing.image.NumpyArrayIterator
                Iterator used to train the model.
        """

        trial_data_generator = tf.keras.preprocessing.image.ImageDataGenerator(
            preprocessing_function=self.preprocessing_function(
                augmentation=augmentation))

        trial_iterator = tf.keras.preprocessing.image.NumpyArrayIterator(
            X, y, trial_data_generator, batch_size=batch_size, shuffle=shuffle,
            sample_weight=None,
            seed=None, data_format=None, save_to_dir=None, save_prefix='',
            save_format='png', subset=None, dtype=None
        )
        return trial_iterator

    def fit(self, X, y, fine_tuning=False, shuffle_before_fit=False,
            augmentation=True, **kwargs):
        """Fit the model. All additional keras parameters of class
        tensorflow.keras.Model will pass through. See keras documentation to
        know what can you do: https://keras.io/api/models/model_training_apis/.

        If no parameters are specified, some default options are set [1]:

            - Epochs: 100 if fine_tuning else 500
            - Batch size: 32 if fine_tuning else 2048
            - Callbacks: self.fine_tuning_callbacks if fine_tuning else
                self.training_callbacks

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        y: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Inception
            [n_observ x n_samples x n_channels x 1]
        fine_tuning: bool
            Set to True to use the default training parameters for fine
            tuning. False by default.
        shuffle_before_fit: bool
            If True, the data will be shuffled before training just once. Note
            that if you use the keras native argument 'shuffle', the data is
            shuffled each epoch.
        kwargs:
            Key-value arguments will be passed to the fit function of the model.
            This way, you can set your own training parameters using keras API.
            See https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
        """
        # Check GPU
        if not tensorflow_integration.check_gpu_acceleration():
            warnings.warn('GPU acceleration is not available. The training '
                          'time is drastically reduced with GPU.')

        # Shuffle the data before fitting
        if shuffle_before_fit:
            X, y = sk_utils.shuffle(X, y)

        # Creation of validation split
        val_split = kwargs['validation_split'] if \
                'validation_split' in kwargs else 0.4

        val_idx = None
        train_idx = None
        for i, label in enumerate(np.unique(y)):
            idx = np.where((y == label))[0]
            np.random.shuffle(idx)
            val_idx_label = int(np.round(len(idx) * val_split))

            val_idx = np.concatenate((val_idx, idx[:val_idx_label]), axis=0) \
                if val_idx is not None else np.array(idx[:val_idx_label])
            train_idx = np.concatenate((train_idx, idx[val_idx_label:]), axis=0) \
                if train_idx is not None else np.array(idx[val_idx_label:])

        # Training parameters
        if not fine_tuning:
            # Rewrite default values
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 500
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 32
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.training_callbacks
        else:
            kwargs['epochs'] = kwargs['epochs'] if \
                'epochs' in kwargs else 100
            kwargs['batch_size'] = kwargs['batch_size'] if \
                'batch_size' in kwargs else 32
            kwargs['callbacks'] = kwargs['callbacks'] if \
                'callbacks' in kwargs else self.fine_tuning_callbacks
        if self.initialized:
            for layer in self.model.layers[:-1]:
                layer.trainable = False
        # Transform data if necessary
        X, y = self.transform_data(X, y)
        # Fit
        with tf.device(tensorflow_integration.get_tf_device_name()):
            # return self.model.fit(X, y, **kwargs)
            return self.model.fit(self.trial_iterator(X[train_idx],
                                                      y[train_idx],
                                                      batch_size=kwargs[
                                                          'batch_size'],
                                                      augmentation=augmentation),
                                  validation_data=(X[val_idx], y[val_idx]),
                                  **kwargs)

    def symmetric_channels(self, X, channels):
        """This function takes a set of channels and puts them in a symmetric
        input needed to apply EEGSym.
        """
        left = list()
        right = list()
        middle = list()
        for channel in channels:
            if channel[-1].isnumeric():
                if int(channel[-1]) % 2 == 0:
                    right.append(channel)
                else:
                    left.append(channel)
            else:
                middle.append(channel)
        ordered_channels = left + middle + right
        index_channels = [channels.index(channel) for channel in
                          ordered_channels]
        return X[:, :, index_channels]

    def predict_proba(self, X):
        """Model prediction scores for the given features.

        Parameters
        ----------
        X: np.ndarray
            Feature matrix. If shape is [n_observ x n_samples x n_channels],
            this matrix will be adapted to the input dimensions of EEG-Sym
            [n_observ x n_samples x n_channels x 1]
        """
        # Transform data if necessary
        X = self.transform_data(X)
        # Predict
        with tf.device(tensorflow_integration.get_tf_device_name()):
            return self.model.predict(X)

    def to_pickleable_obj(self):
        # Parameters
        kwargs = {
            'input_time': self.input_time,
            'fs': self.fs,
            'n_cha': self.n_cha,
            'filters_per_branch': self.filters_per_branch,
            'scales_time': self.scales_time,
            'dropout_rate': self.dropout_rate,
            'activation': self.activation,
            'n_classes': self.n_classes,
            'learning_rate': self.learning_rate,
            'ch_lateral': self.ch_lateral
        }
        weights = self.model.get_weights()
        # Pickleable object
        pickleable_obj = {
            'kwargs': kwargs,
            'weights': weights
        }
        return pickleable_obj

    @classmethod
    def from_pickleable_obj(cls, pickleable_obj):
        model = cls(**pickleable_obj['kwargs'])
        model.model.set_weights(pickleable_obj['weights'])
        return model

    def set_weights(self, weights):
        return self.model.set_weights(weights)

    def get_weights(self):
        return self.model.get_weights()

    def save_weights(self, path):
        return self.model.save_weights(path)

    def load_weights(self, weights_path):
        self.initialized = True
        return self.model.load_weights(weights_path)


import json, os


def configure(sql_host, sql_user, sql_passwd, db_folder):
    """
    This function configures medusa server.

    :param sql_host: sql host
    :type sql_host: string

    :param sql_user: sql user
    :type sql_user: string

    :param sql_passwd: sql password
    :type sql_passwd: string
    """

    # Create folder if not exists
    if not os.path.exists(db_folder):
        os.makedirs(db_folder)

    # Save config json
    config = {
        'sql_host': sql_host,
        'sql_user': sql_user,
        'sql_passwd': sql_passwd,
        'db_folder': db_folder,
    }
    path = os.path.dirname(os.path.abspath(__file__)) + '/config.json'
    with open(path, 'w') as f:
        json.dump(config, f)


def load_config():
    try:
        path = os.path.dirname(os.path.abspath(__file__)) + '/config.json'
        f = open(path, 'r')
        config = json.load(f)
        return config
    except IOError as e:
        raise Exception('Configuration file config.json not found. Use medusa_server configure function')



import numpy as np


def check_dimensions(data):
    """This function checks if data dimensions are [n_epochs x n_samples x
    n_channels] and transforms to it if not"""

    # If dimension is samples
    if len(data.shape) == 1:
        return data[np.newaxis, :, np.newaxis]
    # If dimensions are samples and channels
    elif len(data.shape) == 2:
        return data[np.newaxis, :, :]
    else:
        return data




import threading, time, socket, struct
import bson
from medusa.bci.erp_storage import *
from medusa.bci.erp_converters import *
from medusa.bci.erp_models import *


def train_model(app_id, info, bin_runs):
    # Unpack the data
    data = list()
    for bin_run in bin_runs:
        # Unpack run
        data.append(RCPRun.unpack_run(bin_run))

    # Response content
    resp_content = dict()

    # Train ERP detection model
    if info["erp_model"] == "None":
        pass
    if info['erp_model'] == 'STDModel':
        erp_model_settings = STDModelSettings()
        erp_model = STDModel()
        erp_model.configure(erp_model_settings)
        erp_model.fit_dataset(data)
        erp_target_codes_per_seq, erp_sel_codes_per_seq, erp_acc_per_seq = erp_model.assess_performance(data)
        # Response content
        resp_content['erp_selected_codes_per_seq'] = erp_sel_codes_per_seq,
        resp_content['erp_accuracy_per_seq'] = erp_acc_per_seq
        resp_content['erp_model'] = pickle.dumps(erp_model)
    else:
        raise Exception('Unknown ERP detection method')

    # Control state detection model
    if info["cs_model"] == "None":
        pass
    if info["cs_model"] == "OSRD":
        cs_model_settings = OSRDModelSettings()
        cs_model = OSRDModel()
        cs_model.configure(cs_model_settings)
        cs_model.fit_dataset(data)
        cs_sel_state_per_seq, cs_acc_per_seq = cs_model.assess_performance(data)
        resp_content["cs_selected_state_per_seq"] = cs_sel_state_per_seq
        resp_content["cs_accuracy_per_seq"] = cs_acc_per_seq
        resp_content['cs_model'] = pickle.dumps(erp_model)
    else:
        raise Exception('Unknown control state detection method')

    return resp_content


class ERPSpellerSession(threading.Thread):
    """
        This class creates a new ERP session for processing trials in real time. It assumes that client is using the
         RCP paradigm, returning the selected row and column along with the probability vector. If an asynchrony model
         is provided, it also returns the control state of the user.
    """
    def __init__(self, unpacker, erp_model, async_model=None, session_timeout=1800, ports=None):
        threading.Thread.__init__(self)
        # Initialize the session socket
        self.session_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Bind socket to port
        if ports is None:
            # Random available port
            self.session_socket.bind(('', 0))
        elif isinstance(ports, int):
            # Bind to port specified in ports
            self.session_socket.bind(('', ports))
        elif isinstance(ports, list):
            # Bind to the first port available from those specified in ports
            counter = 0
            for p in ports:
                try:
                    self.session_socket.bind(('', p))
                    break
                except Exception as e:
                    print(str(e))
                    counter += 1
            if counter > len(ports):
                raise Exception("Cannot find a free port within range")
        else:
            raise Exception("Input ports must be an int or list")
        # Get port
        self.port = self.session_socket.getsockname()[1]
        # Parameters
        self.unpacker = unpacker
        self.erp_model = erp_model
        self.async_model = async_model
        # Timer
        self.session_timeout = session_timeout
        self.last_conn_time = None

    def log_message(self, msg):
        """Log an arbitrary message.

        This is used by all other logging functions.  Override
        it if you have specific logging wishes.

        The first argument, FORMAT, is a format string for the
        message to be logged.  If the format string contains
        any % escapes requiring parameters, they should be
        specified as subsequent arguments (it's just like
        printf!).

        The client ip and current date/time are prefixed to
        every message.

        """

        """Return the current time formatted for logging."""
        monthname = [None,
                     'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                     'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
        now = time.time()
        year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
        date = "%02d/%3s/%04d %02d:%02d:%02d" % (
            day, monthname[month], year, hh, mm, ss)

        print("\033[92m%s - - [%s] %s\033[0m" % (self.session_socket.getsockname()[0], date, msg))

    def run(self):
        try:
            # Log
            self.log_message("RCP SESSION %s UP" % str(self.port))
            # Session main loop
            request_close = False
            while not request_close:
                self.session_socket.settimeout(self.session_timeout)
                self.session_socket.listen()
                conn, addr = self.session_socket.accept()
                with conn:
                    binary_data = b''
                    while True:
                        part = conn.recv(1024)
                        if not part:
                            if binary_data == b'request_close':
                                request_close = True
                                resp_content = {
                                    "status": 200,
                                    "msg": "Session shutted down"
                                }
                                bin_resp_content = bson.dumps(resp_content)
                                conn.sendall(bin_resp_content)
                                break
                            try:
                                # Unpack the trial
                                fs, n_cha, eeg, times, nseqs, mat_dims, onsets, codes, trials, mat_indices, sequences = self.unpacker.unpack_trial(binary_data)
                                probs = self.erp_model.process_trial(onsets, times, eeg, fs)
                                if self.async_model is not None:
                                    control_state = self.async_model.process_trial(onsets, times, eeg, fs, nseqs)
                                else:
                                    control_state = 1
                                rowcol = rcp_converter(probs, codes, trials, mat_indices, mat_dims)
                                # Trial processed
                                self.log_message("TRIAL PROCESSED (SESSION = %s, LEN = %s B)" % (str(self.port), len(binary_data)))
                                # Headers
                                resp_content = {
                                    "status": 200,
                                    "matrix": rowcol[0],
                                    "sel_row": rowcol[1],
                                    "sel_col": rowcol[2]
                                }
                                bin_resp_content = bson.dumps(resp_content)
                                conn.sendall(bin_resp_content)
                            except Exception as e:
                                resp_content = {
                                    "status": 500,
                                    "msg": str(e)
                                }
                                bin_resp_content = bson.dumps(resp_content)
                                conn.sendall(bin_resp_content)
                            break
                        binary_data += part
            self.log_message("ERP SESSION %s DOWN" % str(self.port))
            self.session_socket.close()
        except socket.timeout:
            self.log_message("ERP SESSION %s TIMEOUT" % str(self.port))
            self.session_socket.close()



import tensorflow as tf
import numpy as np

def __trans_gpu(W):
    """
    Calculates the transitivity using GPU

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_trans : int
        Global transitivity.
        
    """
    K = tf.reduce_sum(tf.where(W != 0,1,0),axis = 1)
    triples = tf.reduce_sum(tf.math.multiply(K,tf.math.subtract(K,1)))
    triangles = tf.math.pow(W,1/3)
    triangles = tf.linalg.matmul(tf.linalg.matmul(triangles,triangles),triangles)
    triangles = tf.linalg.tensor_diag_part(triangles)
    global_trans = tf.math.divide(tf.reduce_sum(triangles),tf.cast(triples,dtype=tf.float64))
        
    return global_trans


def __trans_cpu(W):
    """
    Calculates the transitivity using CPU

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_trans : int
        Global transitivity.
        
    """
    
    K = np.sum(np.where(W != 0,1,0),axis = 1)
    triples = np.sum(K * (K-1))
    triangles = np.diag(np.linalg.matrix_power(W**(1/3),3))
    global_trans = np.sum(triangles) / triples
            
    return global_trans


def transitivity(W,mode):
    """
    Calculates the transitivity, which is the number of triangles divided by 
    the number of triples.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    global_trans : int
        Global transitivity.

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    if mode == 'CPU':
        global_trans = __trans_cpu(W)
    elif mode == 'GPU':
        global_trans = __trans_gpu(W)
    else:
        raise ValueError('Unknown mode')
        
    return global_trans

# import scipy.io as rmat

# data = rmat.loadmat('D:/OneDrive - Universidad de Valladolid/Scripts/testPython/graphTest.mat')
# W = data['W']
# aa = transitivity(W,'CPU')



import numpy as np
from abc import ABC, abstractmethod


class SignalGenerator(ABC):

    def __init__(self, fs):
        self.fs = fs

    @abstractmethod
    def get_chunk(self, duration, n_channels):
        pass


class SinusoidalSignalGenerator(SignalGenerator):

    """Experimental class for sinusoidal signals generation
    """

    def __init__(self, fs, freqs, noise_type='white',
                 noise_params={'mean': 0, 'sigma': 1}):
        # Check errors
        assert noise_type in (None, 'white'), 'Unknown noise type'
        if noise_type == 'white':
            assert 'mean' in noise_params and 'sigma' in noise_params, \
                'For white noise, you must define mean and sigma noise params'
        # Attributes
        super().__init__(fs)
        self.freqs = freqs
        self.noise_type = noise_type
        self.noise_params = noise_params

    def get_chunk(self, duration, n_channels):
        t = np.linspace(0, duration-(1/self.fs), self.fs*duration)
        chunk = np.zeros((t.shape[0], n_channels))
        # Add sinusoids
        for f in self.freqs:
            s = np.sin(2*np.pi*f*t).reshape(-1, 1)
            chunk += s
        # Add noise
        if self.noise_type == 'white':
            chunk += np.random.random(chunk.shape)
        return chunk



import numpy as np


def reject_noisy_epochs(epochs, signal_mean, signal_std, k=4, n_samp=2,
                        n_cha=1):
    """Simple thresholding method to reject noisy epochs. It discards epochs
    with n_samp samples greater than k*std in n_cha channels

    Parameters
    ----------
     epochs : list or numpy.ndarray
        Epochs of signal with dimensions [n_epochs x samples x channels]
    signal_mean : float
        Mean of the signal
    signal_std : float
        Standard deviation of the signal
    k : float
        Standard deviation multiplier to calculate threshold
    n_samp : int
        Minimum number of samples that have to be over the threshold in each
        epoch to be discarded
    n_cha : int
        Minimum number of channels that have to have n_samples over the
        threshold in each epoch to be discarded

    Returns
    -------
     float
        Percentage of reject epochs in
    numpy.ndarray
        Clean epochs
    numpy.ndarray
        Indexes for rejected epochs. True for discarded epoch
    """

    # Check errors
    if len(epochs.shape) != 3:
        raise Exception('Malformed epochs array. It must be of dimmensions '
                        '[epochs x samples x channels]')
    if signal_std.shape[0] != epochs.shape[2]:
        raise Exception('Array signal_std does not match with epochs size. '
                        'It must have the same number of channels')
    if signal_mean.shape[0] != epochs.shape[2]:
        raise Exception('Array signal_mean does not match with epochs size. '
                        'It must have the same number of channels')

    epochs_abs = np.abs(epochs)
    cmp = epochs_abs > np.abs(signal_mean) + k * signal_std
    idx = np.sum((np.sum(cmp, axis=1) >= n_samp), axis=1) >= n_cha
    pct_rejected = (np.sum(idx) / epochs.shape[0]) * 100
    return pct_rejected, epochs[~idx, :, :], idx
